!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Applied and Computational Mathematics Division                  !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

! This file contains two modules, the grid type module that is very much
! like the 2D grid type module, and a boundary type module that contains
! additional types for handling boundary surfaces

module boundtype_mod

!----------------------------------------------------
! This module contains data structures for the boundary.
!----------------------------------------------------

!----------------------------------------------------
! Other modules used:

use global
!----------------------------------------------------

implicit none
private
public bvert_type, line_type, lineloop_type, surf_type, &
       SURF_PLANAR, SURF_RULED_3, SURF_RULED_4, &
       LINE_LINE, LINE_CIRCLE, LINE_ELLIPSE

!----------------------------------------------------
! The following parameters are defined:

integer, parameter :: SURF_PLANAR  = 1, &
                      SURF_RULED_3 = 2, &
                      SURF_RULED_4 = 3

integer, parameter :: LINE_LINE    = 1, &
                      LINE_CIRCLE  = 2, &
                      LINE_ELLIPSE = 3

!----------------------------------------------------
! The following types are defined:

type bvert_type
   integer :: surface      ! ID of associated surface
   integer :: lines(2)     ! ID of lines containing point, or 0
   real(my_real) :: p1, p2 ! PLANAR: fraction of way from lines(1,2)%point(1)
                           !         to the other end point
                           ! RULED_3: (b1,b2) note b3=1-b1-b2
                           ! RULED_4: (xi,eta)
   type(bvert_type), pointer :: next
end type bvert_type

type line_type
   integer :: type
   integer, pointer :: point(:)
end type line_type

type lineloop_type
   integer, pointer :: line(:)
end type lineloop_type

type surf_type
   integer :: type
   integer, pointer :: lineloop(:)
end type surf_type

end module boundtype_mod

module gridtype_mod

!----------------------------------------------------
! This module contains data structures for the grid.
!----------------------------------------------------

!----------------------------------------------------
! Other modules used:

use global
use hash_mod
use message_passing
use boundtype_mod
use linked_list
!----------------------------------------------------

implicit none
private
public VERTICES_PER_ELEMENT, EDGES_PER_ELEMENT, FACES_PER_ELEMENT, &
       NEIGHBORS_PER_ELEMENT, VERTICES_PER_FACE, EDGES_PER_FACE, &
       MAX_CHILD, ALL_CHILDREN, END_OF_LIST, NO_CHILD, BOUNDARY, &
       GRID_DIM, BASIS_DIM, COMP_DIM, EIGEN_DIM, UNKNOWN_OWNER, &
       point, element_t, face_t, edge_t, vertex_t, grid_type, refine_options, &
       eigen_monitor, element_kind, MAX_TAG, point_type, &
       operator(+), operator(-), operator(*), operator(/), dot_point, &
       operator(.eq.), &
       store_boundary_vertex, restore_boundary_vertex, &
       store_boundary_components, restore_boundary_components, &
       get_vert_soln, get_edge_soln, get_face_soln, get_elem_soln, &
       set_vert_soln, set_edge_soln, set_face_soln, set_elem_soln, &
       copy_vert_soln, copy_edge_soln, copy_face_soln, copy_elem_soln, &
       allocate_vert_soln, allocate_edge_soln, allocate_face_soln, &
       allocate_elem_soln, &
       reallocate_vert_soln, reallocate_edge_soln, reallocate_face_soln, &
       reallocate_elem_soln, &
       deallocate_vert_soln, deallocate_edge_soln, deallocate_face_soln, &
       deallocate_elem_soln, &
       associated_vert_soln, associated_edge_soln, associated_face_soln, &
       associated_elem_soln, &
       nullify_vert_soln, nullify_edge_soln, nullify_face_soln, &
       nullify_elem_soln, &
       size_vert_soln, size_edge_soln, size_face_soln, size_elem_soln, &
       get_vert_exact, get_edge_exact, get_face_exact, get_elem_exact, &
       set_vert_exact, set_edge_exact, set_face_exact, set_elem_exact, &
       copy_vert_exact, copy_edge_exact, copy_face_exact, copy_elem_exact, &
       allocate_vert_exact, allocate_edge_exact, allocate_face_exact, &
       allocate_elem_exact, &
       reallocate_vert_exact, reallocate_edge_exact, reallocate_face_exact, &
       reallocate_elem_exact, &
       deallocate_vert_exact, deallocate_edge_exact, deallocate_face_exact, &
       deallocate_elem_exact, &
       associated_vert_exact, associated_edge_exact, associated_face_exact, &
       associated_elem_exact, &
       nullify_vert_exact, nullify_edge_exact, nullify_face_exact, &
       nullify_elem_exact, &
       size_vert_exact, size_edge_exact, size_face_exact, size_elem_exact, &
       get_vert_oldsoln, get_edge_oldsoln, get_face_oldsoln, get_elem_oldsoln, &
       set_vert_oldsoln, set_edge_oldsoln, set_face_oldsoln, set_elem_oldsoln, &
       copy_vert_oldsoln, copy_edge_oldsoln, copy_face_oldsoln, &
       copy_elem_oldsoln, &
       allocate_vert_oldsoln, allocate_edge_oldsoln, allocate_face_oldsoln, &
       allocate_elem_oldsoln, &
       reallocate_vert_oldsoln, reallocate_edge_oldsoln, &
       reallocate_face_oldsoln, reallocate_elem_oldsoln, &
       deallocate_vert_oldsoln, deallocate_edge_oldsoln, &
       deallocate_face_oldsoln, deallocate_elem_oldsoln, &
       associated_vert_oldsoln, associated_edge_oldsoln, &
       associated_face_oldsoln, associated_elem_oldsoln, &
       nullify_vert_oldsoln, nullify_edge_oldsoln, nullify_face_oldsoln, &
       nullify_elem_oldsoln, &
       size_vert_oldsoln, size_edge_oldsoln, size_face_oldsoln, &
       size_elem_oldsoln

!----------------------------------------------------
! The following interface operators are defined:

interface operator(+)
   module procedure point_plus_point
end interface

interface operator(-)
   module procedure point_minus_point
end interface

interface operator(*)
   module procedure scalar_times_point
end interface

interface operator(/)
   module procedure point_div_scalar
end interface

interface operator(.eq.)
   module procedure point_equals_point
end interface

interface get_vert_soln
   module procedure get_vert_soln_sss, get_vert_soln_ssa, get_vert_soln_ssr, &
                    get_vert_soln_sas, get_vert_soln_saa, get_vert_soln_raa, &
                    get_vert_soln_vss
end interface

interface get_edge_soln
   module procedure get_edge_soln_sss, get_edge_soln_ssa, get_edge_soln_sas, &
                    get_edge_soln_saa, get_edge_soln_ssr, get_edge_soln_ass, &
                    get_edge_soln_asa, get_edge_soln_aaa, get_edge_soln_rss, &
                    get_edge_soln_rrr, get_edge_soln_raaa
end interface

interface get_face_soln
   module procedure get_face_soln_sss, get_face_soln_ssr, get_face_soln_sas, &
                    get_face_soln_saa, get_face_soln_ass, get_face_soln_asa, &
                    get_face_soln_aaa, get_face_soln_rrr, get_face_soln_raaa
end interface

interface get_elem_soln
   module procedure get_elem_soln_sss, get_elem_soln_sas, get_elem_soln_saa, &
                    get_elem_soln_sar, get_elem_soln_ass, get_elem_soln_aaa, &
                    get_elem_soln_rss, get_elem_soln_rrr, get_elem_soln_raaa
end interface

interface set_vert_soln
   module procedure set_vert_soln_sss, set_vert_soln_ssa, set_vert_soln_ssac,&
                    set_vert_soln_ssr, set_vert_soln_saa, set_vert_soln_saac,&
                    set_vert_soln_raa
end interface

interface set_edge_soln
   module procedure set_edge_soln_sss, set_edge_soln_ssr, set_edge_soln_ssac,&
                    set_edge_soln_saa, set_edge_soln_saac,set_edge_soln_ass, &
                    set_edge_soln_asa, set_edge_soln_asac,set_edge_soln_aaa, &
                    set_edge_soln_aaac,set_edge_soln_rss, set_edge_soln_rsac,&
                    set_edge_soln_raaa
end interface

interface set_face_soln
   module procedure set_face_soln_sss, set_face_soln_ssr, set_face_soln_ass, &
                    set_face_soln_asa, set_face_soln_asac,set_face_soln_aaa, &
                    set_face_soln_raac,set_face_soln_aaac,set_face_soln_rsac,&
                    set_face_soln_raaa
end interface

interface set_elem_soln
   module procedure set_elem_soln_sss, set_elem_soln_ssr, set_elem_soln_saa, &
                    set_elem_soln_ass, set_elem_soln_aaa, set_elem_soln_rss, &
                    set_elem_soln_raac,set_elem_soln_aaac,set_elem_soln_raaa
end interface

interface copy_vert_soln
   module procedure copy_vert_soln_ssa, copy_vert_soln_saa, copy_vert_soln_raa
end interface

interface copy_edge_soln
   module procedure copy_edge_soln_asa, copy_edge_soln_aaa, copy_edge_soln_raaa
end interface

interface copy_face_soln
   module procedure copy_face_soln_asa, copy_face_soln_aaa, copy_face_soln_raaa
end interface

interface copy_elem_soln
   module procedure copy_elem_soln_aaa, copy_elem_soln_raaa
end interface

interface get_vert_exact
   module procedure get_vert_exact_sss, get_vert_exact_saa, get_vert_exact_aaa,&
                    get_vert_exact_raa
end interface

interface get_edge_exact
   module procedure get_edge_exact_sss, get_edge_exact_aaa, get_edge_exact_raaa
end interface

interface get_face_exact
   module procedure get_face_exact_sss, get_face_exact_aaa, get_face_exact_raaa
end interface

interface get_elem_exact
   module procedure get_elem_exact_sss, get_elem_exact_aaa, get_elem_exact_raaa
end interface

interface set_vert_exact
   module procedure set_vert_exact_sss, set_vert_exact_saa, set_vert_exact_aaa,&
                    set_vert_exact_raa
end interface

interface set_edge_exact
   module procedure set_edge_exact_sss, set_edge_exact_aaa, &
                    set_edge_exact_aaac,set_edge_exact_raaa
end interface

interface set_face_exact
   module procedure set_face_exact_sss, set_face_exact_aaa, &
                    set_face_exact_aaac,set_face_exact_raaa
end interface

interface set_elem_exact
   module procedure set_elem_exact_sss, set_elem_exact_aaa, &
                    set_elem_exact_aaac,set_elem_exact_raaa
end interface

interface copy_vert_exact
   module procedure copy_vert_exact_saa, copy_vert_exact_raa
end interface

interface copy_edge_exact
   module procedure copy_edge_exact_aaa, copy_edge_exact_raaa
end interface

interface copy_face_exact
   module procedure copy_face_exact_aaa, copy_face_exact_raaa
end interface

interface copy_elem_exact
   module procedure copy_elem_exact_aaa, copy_elem_exact_raaa
end interface

interface get_vert_oldsoln
   module procedure get_vert_oldsoln_sss, get_vert_oldsoln_saa, &
                    get_vert_oldsoln_aaa, get_vert_oldsoln_raa
end interface

interface get_edge_oldsoln
   module procedure get_edge_oldsoln_sss, get_edge_oldsoln_saa, &
                    get_edge_oldsoln_aaa, get_edge_oldsoln_raaa
end interface

interface get_face_oldsoln
   module procedure get_face_oldsoln_sss, get_face_oldsoln_aaa, &
                    get_face_oldsoln_raaa
end interface

interface get_elem_oldsoln
   module procedure get_elem_oldsoln_sss, get_elem_oldsoln_aaa, &
                    get_elem_oldsoln_raaa
end interface

interface set_vert_oldsoln
   module procedure set_vert_oldsoln_sss, set_vert_oldsoln_saa, &
                    set_vert_oldsoln_aaa, set_vert_oldsoln_raa
end interface

interface set_edge_oldsoln
   module procedure set_edge_oldsoln_sss, set_edge_oldsoln_aaa, &
                    set_edge_oldsoln_aaac,set_edge_oldsoln_raaa
end interface

interface set_face_oldsoln
   module procedure set_face_oldsoln_sss, set_face_oldsoln_aaa, &
                    set_face_oldsoln_raaa
end interface

interface set_elem_oldsoln
   module procedure set_elem_oldsoln_sss, set_elem_oldsoln_aaa, &
                    set_elem_oldsoln_aaac,set_elem_oldsoln_raaa
end interface

interface copy_vert_oldsoln
   module procedure copy_vert_oldsoln_saa, copy_vert_oldsoln_raa
end interface

interface copy_edge_oldsoln
   module procedure copy_edge_oldsoln_aaa, copy_edge_oldsoln_raaa
end interface

interface copy_face_oldsoln
   module procedure copy_face_oldsoln_aaa, copy_face_oldsoln_raaa
end interface

interface copy_elem_oldsoln
   module procedure copy_elem_oldsoln_aaa, copy_elem_oldsoln_raaa
end interface

!----------------------------------------------------
! The following parameters are defined:

! values that depend on the kind of elements, polynomial degree, etc.

integer, parameter :: VERTICES_PER_ELEMENT  = 4, &
                      EDGES_PER_ELEMENT     = 6, &
                      FACES_PER_ELEMENT     = 4, &
                      NEIGHBORS_PER_ELEMENT = 4, &
                      VERTICES_PER_FACE     = 3, &
                      EDGES_PER_FACE        = 3, &
                      MAX_CHILD             = 2

! argument for get_child when all children are wanted

integer, private :: i
integer, parameter :: ALL_CHILDREN(MAX_CHILD) = (/ (i,i=1,MAX_CHILD) /)

! flags

integer, parameter :: END_OF_LIST = -10, & ! end of linked list
                      NO_CHILD    = -12    ! child does not exist

! owner of an element is not known

integer, parameter :: UNKNOWN_OWNER = -2

! neighbor or mate of an element when that side is on the boundary

integer, parameter :: BOUNDARY = -6

! elements are triangles

integer, parameter :: element_kind = TETRAHEDRAL_FIXED_ELEMENT

! maximum number of tags for grid entities

integer, parameter :: MAX_TAG = 3

! order of dimensions for the solution component functions

integer, parameter :: GRID_DIM  = 1, &
                      BASIS_DIM = 4, &
                      COMP_DIM  = 2, &
                      EIGEN_DIM = 3

!----------------------------------------------------
! The following types are defined:

type eigen_monitor
   real(my_real), pointer :: eigenvalue(:)
   real(my_real), pointer :: eigensolver_l2_resid(:), eigensolver_errbound(:)
   integer :: ncv, maxit, niter, nconv
end type eigen_monitor

type point
   real(my_real) :: x,y,z
end type point

! This really belongs in boundtype_mod, but then I couldn't use type(point)

type point_type
   type(point) :: coord
end type point_type

type element_t
   type(hash_key) :: gid
! mate is not used in the 3D h-adaptive refinement; instead it is set to
! BOUNDARY so that processing of the mate is always skipped
   type(hash_key) :: mate
! neighbor_hint is the neighbor, an ancestor of the neighbor or a descendant
! of the neighbor that shares the corresponding face
   type(hash_key) :: neighbor_hint(NEIGHBORS_PER_ELEMENT)
   real(my_real) :: weight
   real(my_real) :: work
   real(my_real) :: sp_eta_pred
   real(my_real) :: diameter
   integer :: vertex(VERTICES_PER_ELEMENT)
   integer :: edge(EDGES_PER_ELEMENT)
   integer :: face(FACES_PER_ELEMENT)
   integer :: degree
   integer :: level
   integer :: in, out
   integer :: order(MAX_CHILD)
   integer :: tags(MAX_TAG)
   type(linked_list_integer) :: alt_tags
   integer :: next, previous ! links for available memory list when not in use
                             ! and elements of one level when in use
   integer :: refinement_edge
   integer :: owner
   logical(small_logical) :: isleaf, oldleaf
   logical(small_logical) :: iown ! true if I own all leaves below this element
   logical(small_logical) :: hrefined_unowned, prefined_unowned
   logical(small_logical) :: gmsh_element
   character(len=2) :: type
end type element_t

! there is no linked list of vertices, edges or faces in use.  next is used for
! linked list of free memory and bidirectional pointers between PERIODIC_SLAVEs
! and PERIODIC_MASTERs

type face_t
   type(hash_key) :: gid
   integer :: edge(EDGES_PER_FACE)
   integer :: vertex(VERTICES_PER_FACE)
   integer :: child(2)
   integer :: bmark, degree, assoc_elem, marked_edge, tags(MAX_TAG)
   integer, pointer :: alt_bmark(:)
   integer :: next
   logical(small_logical) :: gmsh_element
end type face_t

type edge_t
   type(hash_key) :: gid
   integer :: vertex(2)
   integer :: child(2)
   integer :: bmark, degree, assoc_elem, tags(MAX_TAG)
   integer, pointer :: alt_bmark(:)
   integer :: next
   logical(small_logical) :: gmsh_element
end type edge_t

type vertex_t
   type(hash_key) :: gid
   type(point) :: coord
   real(my_real) :: bparam
   integer :: bmark, assoc_elem, tags(MAX_TAG)
   integer, pointer :: alt_bmark(:)
   integer :: next
   type(bvert_type), pointer :: boundary_vertex
   logical(small_logical) :: gmsh_element
   type(linked_list_integer) :: node_tris, node_tets
end type vertex_t

! if not solving an eigenproblem, num_eval is 0 and nsoln is system_size.

type grid_type
   type(element_t), pointer :: element(:) => NULL()
   type(face_t), pointer :: face(:) => NULL()
   type(edge_t), pointer :: edge(:) => NULL()
   type(vertex_t), pointer :: vertex(:) => NULL()
   type(hash_table) :: elem_hash, face_hash, edge_hash, vert_hash
   type(point) :: boundbox_min, boundbox_max
! subscripts are lid, system rank, eigenvalue
   real(my_real), pointer :: vertex_solution(:,:,:), vertex_exact(:,:,:), &
                             vertex_oldsoln(:,:,:)
! subscripts are lid, system rank, eigenvalue, basis rank
   real(my_real), pointer :: edge_solution(:,:,:,:), edge_exact(:,:,:,:), &
                             edge_oldsoln(:,:,:,:)
   real(my_real), pointer :: face_solution(:,:,:,:), face_exact(:,:,:,:), &
                             face_oldsoln(:,:,:,:)
   real(my_real), pointer :: elem_solution(:,:,:,:), elem_exact(:,:,:,:), &
                             elem_oldsoln(:,:,:,:)
! subscripts are element, eigenvalue
   real(my_real), pointer :: element_errind(:,:)
   real(my_real), pointer :: eigenprob_l2_resid(:), eigenprob_variance(:)
   real(my_real), pointer :: errest_energy(:), errest_Linf(:), errest_L2(:), &
                             errest_H1(:), errest_H1semi(:), &
                             errest_eigenvalue(:)
   real(my_real) :: max_blen
   real(my_real), pointer :: bp_start(:), bp_finish(:)
   integer, pointer :: element_type(:,:)
   integer, pointer :: face_type(:,:), edge_type(:,:), vertex_type(:,:)
   integer, pointer :: initial_neighbor(:,:) ! (NEIGHBORS_PER_ELEMENT,nelem_init)
   integer, pointer :: head_level_elem(:), head_level_vert(:)
   integer :: next_free_elem, next_free_face, next_free_edge, next_free_vert
   integer :: partition
   integer :: edge_dim_basis, face_dim_basis, elem_dim_basis
   integer :: system_size, num_eval, nsoln
   integer :: nelem, nelem_leaf, nelem_leaf_own, nface, nface_leaf, &
              nface_leaf_own, nedge, nedge_leaf, nedge_leaf_own, nvert, &
              nvert_own, nlev, dof, dof_own, init_nvert, init_nedge, &
              init_nface, init_nelem
   integer :: biggest_vert, biggest_edge, biggest_face, biggest_elem
   integer :: errtype ! really belongs in io_options but that doesn't get
                      ! passed where it is needed
   logical :: errind_up2date, oldsoln_exists, have_true, any_periodic, &
              grid_modified
   character(len=FN_LEN) :: triangle_files
   type(eigen_monitor) :: eigen_results
   type(point_type), pointer :: point(:)
   type(line_type), pointer :: line(:)
   type(lineloop_type), pointer :: lineloop(:)
   type(surf_type), pointer :: surface(:)
   integer :: npoint, nline, nlineloop, nsurface
   logical :: all_surfaces_are_polygonal
   integer, pointer :: floating_tags(:)
   real(my_real), pointer :: floating_total_charge(:)
end type grid_type

type refine_options
   real(my_real) :: inc_factor, reftol, reftol_factor, coarsentol, &
                    coarsentol_factor, recoarsen_tol, term_energy_err, &
                    term_Linf_err, term_H1_err, term_H1_seminorm_err, &
                    term_L2_err, term_eigenvalue_err, tp_gamma, sp_gamma_h, &
                    sp_gamma_p, pbias, min_size
   integer :: error_estimator, reftype, refterm, refmaxloop, recoarsen_limit, &
              edge_rule, hp_strategy, max_vert, max_elem, max_dof, max_lev, &
              max_deg, p_compatibility, max_p_diff
   logical :: derefine, stop_on_maxlev, stop_on_maxdeg, reftol_scale_energy
   real(my_real), pointer :: hp_patch_mult(:,:) => NULL()
   integer, pointer :: hp_patch_add(:,:) => NULL()
end type refine_options

! Undocumented feature to make the grid and current element lid available to
! pde_coefs to, for example, access tags.  But don't tell anyone!
! pde_coefs need to use both global and gridtype_mod.

type(grid_type), pointer, public :: secret_grid

contains

! some operators on points

function point_plus_point(p1,p2)
type(point), intent(in) :: p1, p2
type(point) :: point_plus_point
point_plus_point%x = p1%x + p2%x
point_plus_point%y = p1%y + p2%y
point_plus_point%z = p1%z + p2%z
end function point_plus_point

function point_minus_point(p1,p2)
type(point), intent(in) :: p1, p2
type(point) :: point_minus_point
point_minus_point%x = p1%x - p2%x
point_minus_point%y = p1%y - p2%y
point_minus_point%z = p1%z - p2%z
end function point_minus_point

function scalar_times_point(a,p1)
real(my_real), intent(in) :: a
type(point), intent(in) :: p1
type(point) :: scalar_times_point
scalar_times_point%x = a*p1%x
scalar_times_point%y = a*p1%y
scalar_times_point%z = a*p1%z
end function scalar_times_point

function point_div_scalar(p1,a)
type(point), intent(in) :: p1
real(my_real), intent(in) :: a
type(point) :: point_div_scalar
point_div_scalar%x = p1%x/a
point_div_scalar%y = p1%y/a
point_div_scalar%z = p1%z/a
end function point_div_scalar

function point_equals_point(p1,p2)
type(point), intent(in) :: p1, p2
logical :: point_equals_point
point_equals_point = p1%x == p2%x .and. p1%y == p2%y .and. p1%z == p2%z
end function point_equals_point

function dot_point(p1,p2)
! This routine computes the dot product of two points (meaning the dot
! product of the vectors from the origin to the points)
type(point), intent(in) :: p1, p2
real(my_real) :: dot_point
dot_point = p1%x*p2%x + p1%y*p2%y + p1%z*p2%z
end function dot_point

! subroutines to store and restore the boundary components for phaml_store
! and phaml_restore.

subroutine store_boundary_vertex(grid,vert,unit)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert, unit
character(len=11) :: form
type(bvert_type), pointer :: bvert
inquire(unit,form=form)
select case(form)
case ("UNDEFINED")
   ierr = USER_INPUT_ERROR
   call fatal("File appears unopened in store_boundary_vertex",intlist=(/unit/))
   stop
case ("FORMATTED")
   write(unit,*) 1 ! version number
   bvert => grid%vertex(vert)%boundary_vertex
   do while (associated(bvert))
      write(unit,*) 1
      write(unit,*) bvert%surface, bvert%lines, bvert%p1, bvert%p2
      bvert => bvert%next
   end do
   write(unit,*) 0
case ("UNFORMATTED")
   write(unit) 1 ! version number
   bvert => grid%vertex(vert)%boundary_vertex
   do while (associated(bvert))
      write(unit) 1
      write(unit) bvert%surface, bvert%lines, bvert%p1, bvert%p2
      bvert => bvert%next
   end do
   write(unit) 0
end select
end subroutine store_boundary_vertex

subroutine restore_boundary_vertex(grid,vert,unit)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert, unit
character(len=11) :: form
integer :: version, itemp, astat
type(bvert_type), pointer :: bvert
inquire(unit,form=form)
select case(form)
case ("UNDEFINED")
   ierr = USER_INPUT_ERROR
   call fatal("File appears unopened in restore_boundary_vertex",intlist=(/unit/))
   stop
case ("FORMATTED")
   read(unit,*) version
   select case(version)
   case (1)
      read(unit,*) itemp
      if (itemp == 1) then
         allocate(grid%vertex(vert)%boundary_vertex,stat=astat)
         if (astat /= 0) then
            ierr = ALLOC_FAILED
            call fatal("allocation failed in restore_boundary_vertex")
            stop
         endif
         bvert => grid%vertex(vert)%boundary_vertex
      else
         nullify(grid%vertex(vert)%boundary_vertex)
      endif
      do while(itemp == 1)
         read(unit,*) bvert%surface, bvert%lines, bvert%p1, bvert%p2
         read(unit,*) itemp
         if (itemp == 1) then
            allocate(bvert%next,stat=astat)
            if (astat /= 0) then
               ierr = ALLOC_FAILED
               call fatal("allocation failed in restore_boundary_vertex")
               stop
            endif
            bvert => bvert%next
         else
            nullify(bvert%next)
         endif
      end do
   case default
      call fatal("Unknown version number in restore_boundary_vertex")
      stop
   end select
case ("UNFORMATTED")
   read(unit) version
   select case(version)
   case (1)
      read(unit) itemp
      if (itemp == 1) then
         allocate(grid%vertex(vert)%boundary_vertex,stat=astat)
         if (astat /= 0) then
            ierr = ALLOC_FAILED
            call fatal("allocation failed in restore_boundary_vertex")
            stop
         endif
         bvert => grid%vertex(vert)%boundary_vertex
      else
         nullify(grid%vertex(vert)%boundary_vertex)
      endif
      do while(itemp == 1)
         read(unit) bvert%surface, bvert%lines, bvert%p1, bvert%p2
         read(unit) itemp
         if (itemp == 1) then
            allocate(bvert%next,stat=astat)
            if (astat /= 0) then
               ierr = ALLOC_FAILED
               call fatal("allocation failed in restore_boundary_vertex")
               stop
            endif
            bvert => bvert%next
         else
            nullify(bvert%next)
         endif
      end do
   case default
      call fatal("Unknown version number in restore_boundary_vertex")
      stop
   end select
end select
end subroutine restore_boundary_vertex

subroutine store_boundary_components(grid,unit)
type(grid_type), intent(in) :: grid
integer, intent(in) :: unit
character(len=11) :: form
inquire(unit,form=form)
select case(form)
case ("UNDEFINED")
   ierr = USER_INPUT_ERROR
   call fatal("File appears unopened in store_boundary_components", &
              intlist=(/unit/))
   stop
case ("FORMATTED")
   write(unit,*) 1 ! version number
   write(unit,*) grid%npoint, grid%nline, grid%nlineloop, grid%nsurface
   do i=1,grid%npoint
      write(unit,*) grid%point(i)%coord
   end do
   do i=1,grid%nline
      write(unit,*) size(grid%line(i)%point)
      write(unit,*) grid%line(i)%type,grid%line(i)%point
   end do
   do i=1,grid%nlineloop
      write(unit,*) size(grid%lineloop(i)%line)
      write(unit,*) grid%lineloop(i)%line
   end do
   do i=1,grid%nsurface
      write(unit,*) size(grid%surface(i)%lineloop)
      write(unit,*) grid%surface(i)%type, grid%surface(i)%lineloop
   end do
case ("UNFORMATTED")
   write(unit) 1 ! version number
   write(unit) grid%npoint, grid%nline, grid%nlineloop, grid%nsurface
   do i=1,grid%npoint
      write(unit) grid%point(i)%coord
   end do
   do i=1,grid%nline
      write(unit) size(grid%line(i)%point)
      write(unit) grid%line(i)%type,grid%line(i)%point
   end do
   do i=1,grid%nlineloop
      write(unit) size(grid%lineloop(i)%line)
      write(unit) grid%lineloop(i)%line
   end do
   do i=1,grid%nsurface
      write(unit) size(grid%surface(i)%lineloop)
      write(unit) grid%surface(i)%type, grid%surface(i)%lineloop
   end do
end select
end subroutine store_boundary_components

subroutine restore_boundary_components(grid,unit)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: unit
character(len=11) :: form
integer :: version, itemp, astat
inquire(unit,form=form)
select case(form)
case ("UNDEFINED")
   ierr = USER_INPUT_ERROR
   call fatal("File appears unopened in store_boundary_components", &
              intlist=(/unit/))
   stop
case ("FORMATTED")
   read(unit,*) version
   select case(version)
   case (1)
   read(unit,*) grid%npoint, grid%nline, grid%nlineloop, grid%nsurface
   allocate(grid%point(grid%npoint),grid%line(grid%nline), &
            grid%lineloop(grid%nlineloop),grid%surface(grid%nsurface), &
            stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in restore_boundary_components")
      stop
   endif
   do i=1,grid%npoint
      read(unit,*) grid%point(i)%coord
   end do
   do i=1,grid%nline
      read(unit,*) itemp
      allocate(grid%line(i)%point(itemp),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("allocation failed in restore_boundary_components")
         stop
      endif
      read(unit,*) grid%line(i)%type,grid%line(i)%point
   end do
   do i=1,grid%nlineloop
      read(unit,*) itemp
      allocate(grid%lineloop(i)%line(itemp),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("allocation failed in restore_boundary_components")
         stop
      endif
      read(unit,*) grid%lineloop(i)%line
   end do
   do i=1,grid%nsurface
      read(unit,*) itemp
      allocate(grid%surface(i)%lineloop(itemp),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("allocation failed in restore_boundary_components")
         stop
      endif
      read(unit,*) grid%surface(i)%type, grid%surface(i)%lineloop
   end do
   case default
      call fatal("Unknown version number in restore_boundary_components")
      stop
   end select
case ("UNFORMATTED")
   read(unit) version
   select case(version)
   case (1)
   read(unit) grid%npoint, grid%nline, grid%nlineloop, grid%nsurface
   allocate(grid%point(grid%npoint),grid%line(grid%nline), &
            grid%lineloop(grid%nlineloop),grid%surface(grid%nsurface), &
            stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in restore_boundary_components")
      stop
   endif
   do i=1,grid%npoint
      read(unit) grid%point(i)%coord
   end do
   do i=1,grid%nline
      read(unit) itemp
      allocate(grid%line(i)%point(itemp),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("allocation failed in restore_boundary_components")
         stop
      endif
      read(unit) grid%line(i)%type,grid%line(i)%point
   end do
   do i=1,grid%nlineloop
      read(unit) itemp
      allocate(grid%lineloop(i)%line(itemp),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("allocation failed in restore_boundary_components")
         stop
      endif
      read(unit) grid%lineloop(i)%line
   end do
   do i=1,grid%nsurface
      read(unit) itemp
      allocate(grid%surface(i)%lineloop(itemp),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("allocation failed in restore_boundary_components")
         stop
      endif
      read(unit) grid%surface(i)%type, grid%surface(i)%lineloop
   end do
   case default
      call fatal("Unknown version number in restore_boundary_components")
      stop
   end select
end select
end subroutine restore_boundary_components

! functions for manipulating solution component

! The order of the arguments, subscripts of val, and subscripts of the return
! value are always grid, grid_entity, basis, component, eigenvalue and dim, as
! appropriate.  dim should be GRID_DIM, BASIS_DIM, COMP_DIM or EIGEN_DIM.
! size always checks for allocation and returns 0 if the solution is not
! allocated, so it is not necessary to check allocation before calling size.
! But it is necessary to check allocated or size=0 for get and set in cases
! where it is possible to not be allocated.
! For allocate, the arguments are grid, grid_entity, size_grid, size_basis,
! size_comp and size_eigen, as appropriate.  Other than grid, they are all
! optional.

function get_vert_soln_sss(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert,comp,eigen
real(my_real) :: get_vert_soln_sss
get_vert_soln_sss = grid%vertex_solution(vert,comp,eigen)
end function get_vert_soln_sss

function get_vert_soln_ssa(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert,comp
logical, intent(in) :: eigen
real(my_real) :: get_vert_soln_ssa(size(grid%vertex_solution,dim=EIGEN_DIM))
get_vert_soln_ssa = grid%vertex_solution(vert,comp,:)
end function get_vert_soln_ssa

function get_vert_soln_ssr(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert,comp
integer, intent(in) :: eigen(2)
real(my_real) :: get_vert_soln_ssr(eigen(2)-eigen(1)+1)
get_vert_soln_ssr = grid%vertex_solution(vert,comp,eigen(1):eigen(2))
end function get_vert_soln_ssr

function get_vert_soln_sas(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert,eigen
logical, intent(in) :: comp
real(my_real) :: get_vert_soln_sas(size(grid%vertex_solution,dim=COMP_DIM))
get_vert_soln_sas = grid%vertex_solution(vert,:,eigen)
end function get_vert_soln_sas

function get_vert_soln_saa(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert
logical, intent(in) :: comp,eigen
real(my_real) :: get_vert_soln_saa(size(grid%vertex_solution,dim=COMP_DIM), &
                                   size(grid%vertex_solution,dim=EIGEN_DIM))
get_vert_soln_saa = grid%vertex_solution(vert,:,:)
end function get_vert_soln_saa

function get_vert_soln_raa(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert(2)
logical, intent(in) :: comp,eigen
real(my_real) :: get_vert_soln_raa(vert(2)-vert(1)+1, &
                                   size(grid%vertex_solution,dim=COMP_DIM), &
                                   size(grid%vertex_solution,dim=EIGEN_DIM))
get_vert_soln_raa = grid%vertex_solution(vert(1):vert(2),:,:)
end function get_vert_soln_raa

function get_vert_soln_vss(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: comp,eigen
integer, intent(in) :: vert(:)
real(my_real) :: get_vert_soln_vss(size(vert))
get_vert_soln_vss = grid%vertex_solution(vert,comp,eigen)
end function get_vert_soln_vss

function get_edge_soln_sss(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge,basis,comp,eigen
real(my_real) :: get_edge_soln_sss
get_edge_soln_sss = grid%edge_solution(edge,comp,eigen,basis)
end function get_edge_soln_sss

function get_edge_soln_ssa(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge,basis,comp
logical, intent(in) :: eigen
real(my_real) :: get_edge_soln_ssa(size(grid%edge_solution,dim=EIGEN_DIM))
get_edge_soln_ssa = grid%edge_solution(edge,comp,:,basis)
end function get_edge_soln_ssa

function get_edge_soln_sas(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge,basis,eigen
logical, intent(in) :: comp
real(my_real) :: get_edge_soln_sas(size(grid%edge_solution,dim=COMP_DIM))
get_edge_soln_sas = grid%edge_solution(edge,:,eigen,basis)
end function get_edge_soln_sas

function get_edge_soln_saa(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge,basis
logical, intent(in) :: comp,eigen
real(my_real) :: get_edge_soln_saa(size(grid%edge_solution,dim=COMP_DIM), &
                                   size(grid%edge_solution,dim=EIGEN_DIM))
get_edge_soln_saa = grid%edge_solution(edge,:,:,basis)
end function get_edge_soln_saa

function get_edge_soln_ass(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge
integer, intent(in) :: comp,eigen
logical, intent(in) :: basis
real(my_real) :: get_edge_soln_ass(size(grid%edge_solution,dim=BASIS_DIM))
get_edge_soln_ass = grid%edge_solution(edge,comp,eigen,:)
end function get_edge_soln_ass

function get_edge_soln_asa(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge
integer, intent(in) :: comp
logical, intent(in) :: basis,eigen
integer :: i
real(my_real) :: get_edge_soln_asa(size(grid%edge_solution,dim=BASIS_DIM), &
                                   size(grid%edge_solution,dim=EIGEN_DIM))
do i=1,size(grid%edge_solution,dim=BASIS_DIM)
   get_edge_soln_asa(i,:) = grid%edge_solution(edge,comp,:,i)
end do
end function get_edge_soln_asa

function get_edge_soln_aaa(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_edge_soln_aaa(size(grid%edge_solution,dim=BASIS_DIM), &
                                   size(grid%edge_solution,dim=COMP_DIM), &
                                   size(grid%edge_solution,dim=EIGEN_DIM))
do i=1,size(grid%edge_solution,dim=BASIS_DIM)
   get_edge_soln_aaa(i,:,:) = grid%edge_solution(edge,:,:,i)
end do
end function get_edge_soln_aaa

function get_edge_soln_rss(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge,basis(2),comp,eigen
real(my_real) :: get_edge_soln_rss(basis(2)-basis(1)+1)
get_edge_soln_rss = grid%edge_solution(edge,comp,eigen,basis(1):basis(2))
end function get_edge_soln_rss

function get_edge_soln_ssr(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge,basis,comp,eigen(2)
real(my_real) :: get_edge_soln_ssr(eigen(2)-eigen(1)+1)
get_edge_soln_ssr = grid%edge_solution(edge,comp,eigen(1):eigen(2),basis)
end function get_edge_soln_ssr

function get_edge_soln_rrr(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge,basis(2),comp(2),eigen(2)
integer :: i
real(my_real) :: get_edge_soln_rrr(basis(2)-basis(1)+1, &
                                    comp(2)-comp(1)+1, &
                                    eigen(2)-eigen(1)+1)
do i=basis(1),basis(2)
   get_edge_soln_rrr(i,:,:) = grid%edge_solution(edge,comp(1):comp(2), &
                                                 eigen(1):eigen(2),i)
end do
end function get_edge_soln_rrr

function get_edge_soln_raaa(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge(2)
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_edge_soln_raaa(edge(2)-edge(1)+1, &
                                    size(grid%edge_solution,dim=BASIS_DIM), &
                                    size(grid%edge_solution,dim=COMP_DIM), &
                                    size(grid%edge_solution,dim=EIGEN_DIM))
do i=1,size(grid%edge_solution,dim=BASIS_DIM)
   get_edge_soln_raaa(:,i,:,:) = grid%edge_solution(edge(1):edge(2),:,:,i)
end do
end function get_edge_soln_raaa

function get_face_soln_sss(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face,basis,comp,eigen
real(my_real) :: get_face_soln_sss
get_face_soln_sss = grid%face_solution(face,comp,eigen,basis)
end function get_face_soln_sss

function get_face_soln_ssr(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face,basis,comp,eigen(2)
real(my_real) :: get_face_soln_ssr(eigen(2)-eigen(1)+1)
get_face_soln_ssr = grid%face_solution(face,comp,eigen(1):eigen(2),basis)
end function get_face_soln_ssr

function get_face_soln_sas(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face,basis,eigen
logical, intent(in) :: comp
real(my_real) :: get_face_soln_sas(size(grid%face_solution,dim=COMP_DIM))
get_face_soln_sas = grid%face_solution(face,:,eigen,basis)
end function get_face_soln_sas

function get_face_soln_saa(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face,basis
logical, intent(in) :: comp,eigen
real(my_real) :: get_face_soln_saa(size(grid%face_solution,dim=COMP_DIM), &
                                   size(grid%face_solution,dim=EIGEN_DIM))
get_face_soln_saa = grid%face_solution(face,:,:,basis)
end function get_face_soln_saa

function get_face_soln_ass(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face
integer, intent(in) :: comp,eigen
logical, intent(in) :: basis
real(my_real) :: get_face_soln_ass(size(grid%face_solution,dim=BASIS_DIM))
get_face_soln_ass = grid%face_solution(face,comp,eigen,:)
end function get_face_soln_ass

function get_face_soln_asa(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face
integer, intent(in) :: comp
logical, intent(in) :: basis,eigen
integer :: i
real(my_real) :: get_face_soln_asa(size(grid%face_solution,dim=BASIS_DIM), &
                                   size(grid%face_solution,dim=EIGEN_DIM))
do i=1,size(grid%face_solution,dim=BASIS_DIM)
   get_face_soln_asa(i,:) = grid%face_solution(face,comp,:,i)
end do
end function get_face_soln_asa

function get_face_soln_aaa(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_face_soln_aaa(size(grid%face_solution,dim=BASIS_DIM), &
                                   size(grid%face_solution,dim=COMP_DIM), &
                                   size(grid%face_solution,dim=EIGEN_DIM))
do i=1,size(grid%face_solution,dim=BASIS_DIM)
   get_face_soln_aaa(i,:,:) = grid%face_solution(face,:,:,i)
end do
end function get_face_soln_aaa

function get_face_soln_rrr(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face,basis(2),comp(2),eigen(2)
integer :: i
real(my_real) :: get_face_soln_rrr(basis(2)-basis(1)+1, &
                                   comp(2)-comp(1)+1, &
                                   eigen(2)-eigen(1)+1)
do i=basis(1),basis(2)
   get_face_soln_rrr(i,:,:) = grid%face_solution(face,comp(1):comp(2), &
                                                 eigen(1):eigen(2),i)
end do
end function get_face_soln_rrr

function get_face_soln_raaa(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face(2)
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_face_soln_raaa(face(2)-face(1)+1, &
                                    size(grid%face_solution,dim=BASIS_DIM), &
                                    size(grid%face_solution,dim=COMP_DIM), &
                                    size(grid%face_solution,dim=EIGEN_DIM))
do i=1,size(grid%face_solution,dim=BASIS_DIM)
   get_face_soln_raaa(:,i,:,:) = grid%face_solution(face(1):face(2),:,:,i)
end do
end function get_face_soln_raaa

function get_elem_soln_sss(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem,basis,comp,eigen
real(my_real) :: get_elem_soln_sss
get_elem_soln_sss = grid%elem_solution(elem,comp,eigen,basis)
end function get_elem_soln_sss

function get_elem_soln_sas(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem,basis,eigen
logical, intent(in) :: comp
real(my_real) :: get_elem_soln_sas(size(grid%elem_solution,dim=COMP_DIM))
get_elem_soln_sas = grid%elem_solution(elem,:,eigen,basis)
end function get_elem_soln_sas

function get_elem_soln_saa(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem,basis
logical, intent(in) :: comp,eigen
real(my_real) :: get_elem_soln_saa(size(grid%elem_solution,dim=COMP_DIM), &
                                   size(grid%elem_solution,dim=EIGEN_DIM))
get_elem_soln_saa = grid%elem_solution(elem,:,:,basis)
end function get_elem_soln_saa

function get_elem_soln_sar(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem,basis,eigen(2)
logical, intent(in) :: comp
real(my_real) :: get_elem_soln_sar(size(grid%elem_solution,dim=COMP_DIM), &
                                    eigen(2)-eigen(1)+1)
get_elem_soln_sar = grid%elem_solution(elem,:,eigen(1):eigen(2),basis)
end function get_elem_soln_sar

function get_elem_soln_ass(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem
integer, intent(in) :: comp,eigen
logical, intent(in) :: basis
real(my_real) :: get_elem_soln_ass(size(grid%elem_solution,dim=BASIS_DIM))
get_elem_soln_ass = grid%elem_solution(elem,comp,eigen,:)
end function get_elem_soln_ass

function get_elem_soln_aaa(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_elem_soln_aaa(size(grid%elem_solution,dim=BASIS_DIM), &
                                   size(grid%elem_solution,dim=COMP_DIM), &
                                   size(grid%elem_solution,dim=EIGEN_DIM))
do i=1,size(grid%elem_solution,dim=BASIS_DIM)
   get_elem_soln_aaa(i,:,:) = grid%elem_solution(elem,:,:,i)
end do
end function get_elem_soln_aaa

function get_elem_soln_rss(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem,basis(2),comp,eigen
real(my_real) :: get_elem_soln_rss(basis(2)-basis(1)+1)
get_elem_soln_rss = grid%elem_solution(elem,comp,eigen,basis(1):basis(2))
end function get_elem_soln_rss

function get_elem_soln_rrr(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem,basis(2),comp(2),eigen(2)
integer :: i
real(my_real) :: get_elem_soln_rrr(basis(2)-basis(1)+1, &
                                   comp(2)-comp(1)+1, &
                                   eigen(2)-eigen(1)+1)
do i=basis(1),basis(2)
   get_elem_soln_rrr(i,:,:) = grid%elem_solution(elem,comp(1):comp(2), &
                                                 eigen(1):eigen(2),i)
end do
end function get_elem_soln_rrr

function get_elem_soln_raaa(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem(2)
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_elem_soln_raaa(elem(2)-elem(1)+1, &
                                    size(grid%elem_solution,dim=BASIS_DIM), &
                                    size(grid%elem_solution,dim=COMP_DIM), &
                                    size(grid%elem_solution,dim=EIGEN_DIM))
do i=1,size(grid%elem_solution,dim=BASIS_DIM)
   get_elem_soln_raaa(:,i,:,:) = grid%elem_solution(elem(1):elem(2),:,:,i)
end do
end function get_elem_soln_raaa

subroutine set_vert_soln_sss(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert,comp,eigen
real(my_real), intent(in) :: val
grid%vertex_solution(vert,comp,eigen) = val
end subroutine set_vert_soln_sss

subroutine set_vert_soln_ssa(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert,comp
logical, intent(in) :: eigen
real(my_real), intent(in) :: val(:)
grid%vertex_solution(vert,comp,:) = val
end subroutine set_vert_soln_ssa

subroutine set_vert_soln_ssac(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert,comp
logical, intent(in) :: eigen
real(my_real), intent(in) :: val
grid%vertex_solution(vert,comp,:) = val
end subroutine set_vert_soln_ssac

subroutine set_vert_soln_ssr(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert,comp
integer, intent(in) :: eigen(2)
real(my_real), intent(in) :: val(:)
grid%vertex_solution(vert,comp,eigen(1):eigen(2)) = val
end subroutine set_vert_soln_ssr

subroutine set_vert_soln_saa(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val(:,:)
grid%vertex_solution(vert,:,:) = val
end subroutine set_vert_soln_saa

subroutine set_vert_soln_saac(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val
grid%vertex_solution(vert,:,:) = val
end subroutine set_vert_soln_saac

subroutine set_vert_soln_raa(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert(2)
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val(:,:,:)
grid%vertex_solution(vert(1):vert(2),:,:) = val
end subroutine set_vert_soln_raa

subroutine set_edge_soln_sss(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%edge_solution(edge,comp,eigen,basis) = val
end subroutine set_edge_soln_sss

subroutine set_edge_soln_ssac(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: basis,comp
logical, intent(in) :: eigen
real(my_real), intent(in) :: val
grid%edge_solution(edge,comp,:,basis) = val
end subroutine set_edge_soln_ssac

subroutine set_edge_soln_ssr(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: basis,comp,eigen(2)
real(my_real), intent(in) :: val(:)
grid%edge_solution(edge,comp,eigen(1):eigen(2),basis) = val
end subroutine set_edge_soln_ssr

subroutine set_edge_soln_saa(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: basis
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val(:,:)
grid%edge_solution(edge,:,:,basis) = val
end subroutine set_edge_soln_saa

subroutine set_edge_soln_saac(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: basis
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val
grid%edge_solution(edge,:,:,basis) = val
end subroutine set_edge_soln_saac

subroutine set_edge_soln_aaa(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%edge_solution(edge,:,:,i) = val(i,:,:)
end do
end subroutine set_edge_soln_aaa

subroutine set_edge_soln_aaac(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%edge_solution(edge,:,:,:) = val
end subroutine set_edge_soln_aaac

subroutine set_edge_soln_ass(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: comp,eigen
logical, intent(in) :: basis
real(my_real), intent(in) :: val(:)
grid%edge_solution(edge,comp,eigen,:) = val
end subroutine set_edge_soln_ass

subroutine set_edge_soln_asa(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: comp
logical, intent(in) :: basis,eigen
real(my_real), intent(in) :: val(:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%edge_solution(edge,comp,:,i) = val(i,:)
end do
end subroutine set_edge_soln_asa

subroutine set_edge_soln_asac(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: comp
logical, intent(in) :: basis,eigen
real(my_real), intent(in) :: val
grid%edge_solution(edge,comp,:,:) = val
end subroutine set_edge_soln_asac

subroutine set_edge_soln_rss(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: basis(2),comp,eigen
real(my_real), intent(in) :: val(:)
grid%edge_solution(edge,comp,eigen,basis(1):basis(2)) = val
end subroutine set_edge_soln_rss

subroutine set_edge_soln_rsac(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: basis(2),comp
logical, intent(in) :: eigen
real(my_real), intent(in) :: val
grid%edge_solution(edge,comp,:,basis(1):basis(2)) = val
end subroutine set_edge_soln_rsac

subroutine set_edge_soln_raaa(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge(2)
logical, intent(in) :: basis, comp, eigen
real(my_real), intent(in) :: val(:,:,:,:)
integer :: i
do i=1,size(val,dim=2)
   grid%edge_solution(edge(1):edge(2),:,:,i) = val(:,i,:,:)
end do
end subroutine set_edge_soln_raaa

subroutine set_face_soln_sss(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
integer, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%face_solution(face,comp,eigen,basis) = val
end subroutine set_face_soln_sss

subroutine set_face_soln_ssr(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
integer, intent(in) :: basis,comp,eigen(2)
real(my_real), intent(in) :: val(:)
grid%face_solution(face,comp,eigen(1):eigen(2),basis) = val
end subroutine set_face_soln_ssr

subroutine set_face_soln_ass(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
integer, intent(in) :: comp,eigen
logical, intent(in) :: basis
real(my_real), intent(in) :: val(:)
grid%face_solution(face,comp,eigen,:) = val
end subroutine set_face_soln_ass

subroutine set_face_soln_asa(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
integer, intent(in) :: comp
logical, intent(in) :: basis,eigen
real(my_real), intent(in) :: val(:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%face_solution(face,comp,:,i) = val(i,:)
end do
end subroutine set_face_soln_asa

subroutine set_face_soln_asac(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
integer, intent(in) :: comp
logical, intent(in) :: basis,eigen
real(my_real), intent(in) :: val
grid%face_solution(face,comp,:,:) = val
end subroutine set_face_soln_asac

subroutine set_face_soln_aaa(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%face_solution(face,:,:,i) = val(i,:,:)
end do
end subroutine set_face_soln_aaa

subroutine set_face_soln_raac(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
integer, intent(in) :: basis(2)
logical, intent(in) :: comp, eigen
real(my_real), intent(in) :: val
grid%face_solution(face,:,:,basis(1):basis(2)) = val
end subroutine set_face_soln_raac

subroutine set_face_soln_aaac(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%face_solution(face,:,:,:) = val
end subroutine set_face_soln_aaac

subroutine set_face_soln_rsac(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
integer, intent(in) :: basis(2),comp
logical, intent(in) :: eigen
real(my_real), intent(in) :: val
grid%face_solution(face,comp,:,basis(1):basis(2)) = val
end subroutine set_face_soln_rsac

subroutine set_face_soln_raaa(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face(2)
logical, intent(in) :: basis, comp, eigen
real(my_real), intent(in) :: val(:,:,:,:)
integer :: i
do i=1,size(val,dim=2)
   grid%face_solution(face(1):face(2),:,:,i) = val(:,i,:,:)
end do
end subroutine set_face_soln_raaa

subroutine set_elem_soln_sss(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
integer, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%elem_solution(elem,comp,eigen,basis) = val
end subroutine set_elem_soln_sss

subroutine set_elem_soln_ssr(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
integer, intent(in) :: basis,comp,eigen(2)
real(my_real), intent(in) :: val(:)
grid%elem_solution(elem,comp,eigen(1):eigen(2),basis) = val
end subroutine set_elem_soln_ssr

subroutine set_elem_soln_saa(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
integer, intent(in) :: basis
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val(:,:)
grid%elem_solution(elem,:,:,basis) = val
end subroutine set_elem_soln_saa

subroutine set_elem_soln_ass(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
integer, intent(in) :: comp,eigen
logical, intent(in) :: basis
real(my_real), intent(in) :: val(:)
grid%elem_solution(elem,comp,eigen,:) = val
end subroutine set_elem_soln_ass

subroutine set_elem_soln_aaa(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%elem_solution(elem,:,:,i) = val(i,:,:)
end do
end subroutine set_elem_soln_aaa

subroutine set_elem_soln_rss(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
integer, intent(in) :: basis(2),comp,eigen
real(my_real), intent(in) :: val(:)
grid%elem_solution(elem,comp,eigen,basis(1):basis(2)) = val
end subroutine set_elem_soln_rss

subroutine set_elem_soln_raac(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
integer, intent(in) :: basis(2)
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val
grid%elem_solution(elem,:,:,basis(1):basis(2)) = val
end subroutine set_elem_soln_raac

subroutine set_elem_soln_aaac(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%elem_solution(elem,:,:,:) = val
end subroutine set_elem_soln_aaac

subroutine set_elem_soln_raaa(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem(2)
logical, intent(in) :: basis, comp, eigen
integer :: i
real(my_real), intent(in) :: val(:,:,:,:)
do i=1,size(val,dim=2)
   grid%elem_solution(elem(1):elem(2),:,:,i) = val(:,i,:,:)
end do
end subroutine set_elem_soln_raaa

subroutine copy_vert_soln_ssa(from_grid,from_vert,to_grid,to_vert,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_vert, to_vert, comp
logical, intent(in) :: eigen
to_grid%vertex_solution(to_vert,comp,:) = &
   from_grid%vertex_solution(from_vert,comp,:)
end subroutine copy_vert_soln_ssa

subroutine copy_vert_soln_saa(from_grid,from_vert,to_grid,to_vert,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_vert, to_vert
logical, intent(in) :: comp, eigen
to_grid%vertex_solution(to_vert,:,:) = from_grid%vertex_solution(from_vert,:,:)
end subroutine copy_vert_soln_saa

subroutine copy_vert_soln_raa(from_grid,to_grid,vert,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: vert(2)
logical, intent(in) :: comp, eigen
to_grid%vertex_solution(vert(1):vert(2),:,:) = &
   from_grid%vertex_solution(vert(1):vert(2),:,:)
end subroutine copy_vert_soln_raa

subroutine copy_edge_soln_asa(from_grid,from_edge,to_grid,to_edge,basis,comp, &
                              eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_edge, to_edge, comp
logical, intent(in) :: basis, eigen
to_grid%edge_solution(to_edge,:,comp,:) = &
   from_grid%edge_solution(from_edge,:,comp,:)
end subroutine copy_edge_soln_asa

subroutine copy_edge_soln_aaa(from_grid,from_edge,to_grid,to_edge,basis,comp, &
                              eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_edge, to_edge
logical, intent(in) :: basis, comp, eigen
to_grid%edge_solution(to_edge,:,:,:) = from_grid%edge_solution(from_edge,:,:,:)
end subroutine copy_edge_soln_aaa

subroutine copy_edge_soln_raaa(from_grid,to_grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: edge(2)
logical, intent(in) :: basis, comp, eigen
to_grid%edge_solution(edge(1):edge(2),:,:,:) = &
   from_grid%edge_solution(edge(1):edge(2),:,:,:)
end subroutine copy_edge_soln_raaa

subroutine copy_face_soln_asa(from_grid,from_face,to_grid,to_face,basis,comp, &
                              eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_face, to_face, comp
logical, intent(in) :: basis, eigen
to_grid%face_solution(to_face,:,comp,:) = &
   from_grid%face_solution(from_face,:,comp,:)
end subroutine copy_face_soln_asa

subroutine copy_face_soln_aaa(from_grid,from_face,to_grid,to_face,basis,comp, &
                              eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_face, to_face
logical, intent(in) :: basis, comp, eigen
to_grid%face_solution(to_face,:,:,:) = from_grid%face_solution(from_face,:,:,:)
end subroutine copy_face_soln_aaa

subroutine copy_face_soln_raaa(from_grid,to_grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: face(2)
logical, intent(in) :: basis, comp, eigen
to_grid%face_solution(face(1):face(2),:,:,:) = &
   from_grid%face_solution(face(1):face(2),:,:,:)
end subroutine copy_face_soln_raaa

subroutine copy_elem_soln_aaa(from_grid,from_elem,to_grid,to_elem,basis,comp, &
                              eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_elem, to_elem
logical, intent(in) :: basis, comp, eigen
to_grid%elem_solution(to_elem,:,:,:) = from_grid%elem_solution(from_elem,:,:,:)
end subroutine copy_elem_soln_aaa

subroutine copy_elem_soln_raaa(from_grid,to_grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: elem(2)
logical, intent(in) :: basis, comp, eigen
to_grid%elem_solution(elem(1):elem(2),:,:,:) = &
   from_grid%elem_solution(elem(1):elem(2),:,:,:)
end subroutine copy_elem_soln_raaa

subroutine allocate_vert_soln(grid,vert,size_grid,size_comp,size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: vert, size_grid, size_comp, size_eigen
integer :: s_comp, s_eigen, astat
if (present(vert)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("vert cannot be present in allocate_vert_soln")
   stop
endif
if (.not. present(size_grid)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("size_grid must be present in allocate_vert_soln")
   stop
endif
if (present(size_comp)) then
   s_comp = size_comp
else
   s_comp = grid%system_size
endif
if (present(size_eigen)) then
   s_eigen = size_eigen
else
   s_eigen = max(1,grid%num_eval)
endif
if (associated(grid%vertex_solution)) deallocate(grid%vertex_solution)
if (size_grid==0 .or. s_comp==0 .or. s_eigen==0) then
   nullify(grid%vertex_solution)
else
   allocate(grid%vertex_solution(size_grid,s_comp,s_eigen),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in allocate_vert_soln")
      stop
   endif
endif
end subroutine allocate_vert_soln

subroutine allocate_edge_soln(grid,edge,size_grid,size_basis,size_comp, &
                              size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, astat
if (present(edge)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in allocate_edge_soln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%edge_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (associated(grid%edge_solution)) deallocate(grid%edge_solution)
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      nullify(grid%edge_solution)
   else
      allocate(grid%edge_solution(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in allocate_edge_soln")
         stop
      endif
   endif
endif
end subroutine allocate_edge_soln

subroutine allocate_face_soln(grid,face,size_grid,size_basis,size_comp, &
                              size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, astat
if (present(face)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in allocate_face_soln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%face_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (associated(grid%face_solution)) deallocate(grid%face_solution)
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      nullify(grid%face_solution)
   else
      allocate(grid%face_solution(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in allocate_face_soln")
         stop
      endif
   endif
endif
end subroutine allocate_face_soln

subroutine allocate_elem_soln(grid,elem,size_grid,size_basis,size_comp, &
                              size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, astat
if (present(elem)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in allocate_elem_soln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%elem_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (associated(grid%elem_solution)) deallocate(grid%elem_solution)
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      nullify(grid%elem_solution)
   else
      allocate(grid%elem_solution(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in allocate_elem_soln")
         stop
      endif
   endif
endif
end subroutine allocate_elem_soln

subroutine reallocate_vert_soln(grid,vert,size_grid,size_comp,size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: vert, size_grid, size_comp, size_eigen
integer :: s_comp, s_eigen, ncopy, ncopyss, ncopyev, astat
real(my_real), pointer :: temp(:,:,:)
if (present(vert)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("vert must not be present in reallocate_vert_soln")
   stop
endif
if (.not. present(size_grid)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("size_grid must be present in reallocate_vert_soln")
   stop
endif
if (present(size_comp)) then
   s_comp = size_comp
else
   s_comp = grid%system_size
endif
if (present(size_eigen)) then
   s_eigen = size_eigen
else
   s_eigen = max(1,grid%num_eval)
endif
if (size_grid==0 .or. s_comp==0 .or. s_eigen==0) then
   if (associated(grid%vertex_solution)) deallocate(grid%vertex_solution)
   nullify(grid%vertex_solution)
else
   allocate(temp(size_grid,s_comp,s_eigen),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in reallocate_vert_soln")
      stop
   endif
   temp(1:min(size_grid,grid%biggest_vert),:,:) = 0.0_my_real
   if (associated(grid%vertex_solution)) then
      ncopy = min(size_grid,size_vert_soln(grid,dim=GRID_DIM))
      ncopyss = min(s_comp,size_vert_soln(grid,dim=COMP_DIM))
      ncopyev = min(s_eigen,size_vert_soln(grid,dim=EIGEN_DIM))
      if (ncopy > 0) then
         temp(1:ncopy,1:ncopyss,1:ncopyev) = &
              grid%vertex_solution(1:ncopy,1:ncopyss,1:ncopyev)
      endif
      deallocate(grid%vertex_solution)
   endif
   grid%vertex_solution => temp
   nullify(temp)
endif
end subroutine reallocate_vert_soln

subroutine reallocate_edge_soln(grid,edge,size_grid,size_basis,size_comp, &
                                size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, ncopy, ncopyss, ncopyev, ncopyb, astat
real(my_real), pointer :: temp(:,:,:,:)
if (present(edge)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in reallocate_edge_soln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%edge_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      if (associated(grid%edge_solution)) deallocate(grid%edge_solution)
      nullify(grid%edge_solution)
   else
      allocate(temp(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in reallocate_edge_soln")
         stop
      endif
      temp(1:min(size_grid,grid%biggest_edge),:,:,:) = 0.0_my_real
      if (associated(grid%edge_solution)) then
         ncopy = min(size_grid,size_edge_soln(grid,dim=GRID_DIM))
         ncopyb  = min(s_basis,size_edge_soln(grid,dim=BASIS_DIM))
         ncopyss = min(s_comp,size_edge_soln(grid,dim=COMP_DIM))
         ncopyev = min(s_eigen,size_edge_soln(grid,dim=EIGEN_DIM))
         if (ncopy > 0 .and. ncopyb > 0) then
            temp(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb) = &
               grid%edge_solution(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb)
         endif
         deallocate(grid%edge_solution)
      endif
      grid%edge_solution => temp
      nullify(temp)
   endif
endif
end subroutine reallocate_edge_soln

subroutine reallocate_face_soln(grid,face,size_grid,size_basis,size_comp, &
                                size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, ncopy, ncopyss, ncopyev, ncopyb, astat
real(my_real), pointer :: temp(:,:,:,:)
if (present(face)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in reallocate_face_soln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%face_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      if (associated(grid%face_solution)) deallocate(grid%face_solution)
      nullify(grid%face_solution)
   else
      allocate(temp(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in reallocate_face_soln")
         stop
      endif
      temp(1:min(size_grid,grid%biggest_face),:,:,:) = 0.0_my_real
      if (associated(grid%face_solution)) then
         ncopy = min(size_grid,size_face_soln(grid,dim=GRID_DIM))
         ncopyb  = min(s_basis,size_face_soln(grid,dim=BASIS_DIM))
         ncopyss = min(s_comp,size_face_soln(grid,dim=COMP_DIM))
         ncopyev = min(s_eigen,size_face_soln(grid,dim=EIGEN_DIM))
         if (ncopy > 0 .and. ncopyb > 0) then
            temp(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb) = &
               grid%face_solution(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb)
         endif
         deallocate(grid%face_solution)
      endif
      grid%face_solution => temp
      nullify(temp)
   endif
endif
end subroutine reallocate_face_soln

subroutine reallocate_elem_soln(grid,elem,size_grid,size_basis,size_comp, &
                                size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, ncopy, ncopyss, ncopyev, ncopyb, astat
real(my_real), pointer :: temp(:,:,:,:)
if (present(elem)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in reallocate_elem_soln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%elem_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      if (associated(grid%elem_solution)) deallocate(grid%elem_solution)
      nullify(grid%elem_solution)
   else
      allocate(temp(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in reallocate_elem_soln")
         stop
      endif
      temp(1:min(size_grid,grid%biggest_elem),:,:,:) = 0.0_my_real
      if (associated(grid%elem_solution)) then
         ncopy = min(size_grid,size_elem_soln(grid,dim=GRID_DIM))
         ncopyb  = min(s_basis,size_elem_soln(grid,dim=BASIS_DIM))
         ncopyss = min(s_comp,size_elem_soln(grid,dim=COMP_DIM))
         ncopyev = min(s_eigen,size_elem_soln(grid,dim=EIGEN_DIM))
         if (ncopy > 0 .and. ncopyb > 0) then
            temp(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb) = &
               grid%elem_solution(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb)
         endif
         deallocate(grid%elem_solution)
      endif
      grid%elem_solution => temp
      nullify(temp)
   endif
endif
end subroutine reallocate_elem_soln

subroutine deallocate_vert_soln(grid)
type(grid_type), intent(inout) :: grid
deallocate(grid%vertex_solution)
end subroutine deallocate_vert_soln

subroutine deallocate_edge_soln(grid,edge)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge
if (present(edge)) then
! do nothing
else
   deallocate(grid%edge_solution)
endif
end subroutine deallocate_edge_soln

subroutine deallocate_face_soln(grid,face)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face
if (present(face)) then
! do nothing
else
   deallocate(grid%face_solution)
endif
end subroutine deallocate_face_soln

subroutine deallocate_elem_soln(grid,elem)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem
if (present(elem)) then
! do nothing
else
   deallocate(grid%elem_solution)
endif
end subroutine deallocate_elem_soln

function associated_vert_soln(grid)
type(grid_type), intent(in) :: grid
logical :: associated_vert_soln
associated_vert_soln = associated(grid%vertex_solution)
end function associated_vert_soln

function associated_edge_soln(grid,edge)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: edge
logical :: associated_edge_soln
if (present(edge)) then
   if (associated(grid%edge_solution)) then
      associated_edge_soln = edge <= size(grid%edge_solution,dim=GRID_DIM)
   else
      associated_edge_soln = .false.
   endif
else
   associated_edge_soln = associated(grid%edge_solution)
endif
end function associated_edge_soln

function associated_face_soln(grid,face)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: face
logical :: associated_face_soln
if (present(face)) then
   if (associated(grid%face_solution)) then
      associated_face_soln = face <= size(grid%face_solution,dim=GRID_DIM)
   else
      associated_face_soln = .false.
   endif
else
   associated_face_soln = associated(grid%face_solution)
endif
end function associated_face_soln

function associated_elem_soln(grid,elem)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: elem
logical :: associated_elem_soln
if (present(elem)) then
   if (associated(grid%elem_solution)) then
      associated_elem_soln = elem <= size(grid%elem_solution,dim=GRID_DIM)
   else
      associated_elem_soln = .false.
   endif
else
   associated_elem_soln = associated(grid%elem_solution)
endif
end function associated_elem_soln

subroutine nullify_vert_soln(grid)
type(grid_type), intent(inout) :: grid
nullify(grid%vertex_solution)
end subroutine nullify_vert_soln

subroutine nullify_edge_soln(grid,edge)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge
if (present(edge)) then
! do nothing
else
   nullify(grid%edge_solution)
endif
end subroutine nullify_edge_soln

subroutine nullify_face_soln(grid,face)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face
if (present(face)) then
! do nothing
else
   nullify(grid%face_solution)
endif
end subroutine nullify_face_soln

subroutine nullify_elem_soln(grid,elem)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem
if (present(elem)) then
! do nothing
else
   nullify(grid%elem_solution)
endif
end subroutine nullify_elem_soln

function size_vert_soln(grid,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: dim
integer :: size_vert_soln
integer :: copydim
if (associated_vert_soln(grid)) then
   if (present(dim)) then
      copydim = dim
      size_vert_soln = size(grid%vertex_solution,dim=copydim)
   else
      size_vert_soln = size(grid%vertex_solution)
   endif
else
   size_vert_soln = 0
endif
end function size_vert_soln

function size_edge_soln(grid,edge,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: edge
integer, intent(in), optional :: dim
integer :: size_edge_soln
integer :: copydim
if (associated_edge_soln(grid,edge)) then
   if (present(edge)) then
      if (present(dim)) then
         copydim = dim
         size_edge_soln = size(grid%edge_solution,dim=copydim)
      else
         size_edge_soln = size(grid%edge_solution,dim=BASIS_DIM) * &
                          size(grid%edge_solution,dim=COMP_DIM) * &
                          size(grid%edge_solution,dim=EIGEN_DIM)
      endif
   else
      if (present(dim)) then
         copydim = dim
         size_edge_soln = size(grid%edge_solution,dim=copydim)
      else
         size_edge_soln = size(grid%edge_solution)
      endif
   endif
else
   size_edge_soln = 0
endif
end function size_edge_soln

function size_face_soln(grid,face,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: face
integer, intent(in), optional :: dim
integer :: size_face_soln
integer :: copydim
if (associated_face_soln(grid,face)) then
   if (present(face)) then
      if (present(dim)) then
         copydim = dim
         size_face_soln = size(grid%face_solution,dim=copydim)
      else
         size_face_soln = size(grid%face_solution,dim=BASIS_DIM) * &
                          size(grid%face_solution,dim=COMP_DIM) * &
                          size(grid%face_solution,dim=EIGEN_DIM)
      endif
   else
      if (present(dim)) then
         copydim = dim
         size_face_soln = size(grid%face_solution,dim=copydim)
      else
         size_face_soln = size(grid%face_solution)
      endif
   endif
else
   size_face_soln = 0
endif
end function size_face_soln

function size_elem_soln(grid,elem,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: elem
integer, intent(in), optional :: dim
integer :: size_elem_soln
integer :: copydim
if (associated_elem_soln(grid,elem)) then
   if (present(elem)) then
      if (present(dim)) then
         copydim = dim
         size_elem_soln = size(grid%elem_solution,dim=copydim)
      else
         size_elem_soln = size(grid%elem_solution,dim=BASIS_DIM) * &
                          size(grid%elem_solution,dim=COMP_DIM) * &
                          size(grid%elem_solution,dim=EIGEN_DIM)
      endif
   else
      if (present(dim)) then
         copydim = dim
         size_elem_soln = size(grid%elem_solution,dim=copydim)
      else
         size_elem_soln = size(grid%elem_solution)
      endif
   endif
else
   size_elem_soln = 0
endif
end function size_elem_soln

function get_vert_exact_sss(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert,comp,eigen
real(my_real) :: get_vert_exact_sss
get_vert_exact_sss = grid%vertex_exact(vert,comp,eigen)
end function get_vert_exact_sss

function get_vert_exact_saa(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert
logical, intent(in) :: comp,eigen
real(my_real) :: get_vert_exact_saa(size(grid%vertex_exact,dim=COMP_DIM), &
                                    size(grid%vertex_exact,dim=EIGEN_DIM))
get_vert_exact_saa = grid%vertex_exact(vert,:,:)
end function get_vert_exact_saa

function get_vert_exact_aaa(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
logical, intent(in) :: vert,comp,eigen
real(my_real) :: get_vert_exact_aaa(size(grid%vertex_exact,dim=GRID_DIM), &
                                    size(grid%vertex_exact,dim=COMP_DIM), &
                                    size(grid%vertex_exact,dim=EIGEN_DIM))
get_vert_exact_aaa = grid%vertex_exact(:,:,:)
end function get_vert_exact_aaa

function get_vert_exact_raa(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert(2)
logical, intent(in) :: comp,eigen
real(my_real) :: get_vert_exact_raa(vert(2)-vert(1)+1, &
                                    size(grid%vertex_exact,dim=COMP_DIM), &
                                    size(grid%vertex_exact,dim=EIGEN_DIM))
get_vert_exact_raa = grid%vertex_exact(vert(1):vert(2),:,:)
end function get_vert_exact_raa

function get_edge_exact_sss(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge,basis,comp,eigen
real(my_real) :: get_edge_exact_sss
get_edge_exact_sss = grid%edge_exact(edge,comp,eigen,basis)
end function get_edge_exact_sss

function get_edge_exact_aaa(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_edge_exact_aaa(size(grid%edge_exact,dim=BASIS_DIM), &
                                    size(grid%edge_exact,dim=COMP_DIM), &
                                    size(grid%edge_exact,dim=EIGEN_DIM))
do i=1,size(grid%edge_exact,dim=BASIS_DIM)
   get_edge_exact_aaa(i,:,:) = grid%edge_exact(edge,:,:,i)
end do
end function get_edge_exact_aaa

function get_edge_exact_raaa(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge(2)
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_edge_exact_raaa(edge(2)-edge(1)+1, &
                                     size(grid%edge_exact,dim=BASIS_DIM), &
                                     size(grid%edge_exact,dim=COMP_DIM), &
                                     size(grid%edge_exact,dim=EIGEN_DIM))
do i=1,size(grid%edge_exact,dim=BASIS_DIM)
   get_edge_exact_raaa(:,i,:,:) = grid%edge_exact(edge(1):edge(2),:,:,i)
end do
end function get_edge_exact_raaa

function get_face_exact_sss(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face,basis,comp,eigen
real(my_real) :: get_face_exact_sss
get_face_exact_sss = grid%face_exact(face,comp,eigen,basis)
end function get_face_exact_sss

function get_face_exact_aaa(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_face_exact_aaa(size(grid%face_exact,dim=BASIS_DIM), &
                                    size(grid%face_exact,dim=COMP_DIM), &
                                    size(grid%face_exact,dim=EIGEN_DIM))
do i=1,size(grid%face_exact,dim=BASIS_DIM)
   get_face_exact_aaa(i,:,:) = grid%face_exact(face,:,:,i)
end do
end function get_face_exact_aaa

function get_face_exact_raaa(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face(2)
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_face_exact_raaa(face(2)-face(1)+1, &
                                     size(grid%face_exact,dim=BASIS_DIM), &
                                     size(grid%face_exact,dim=COMP_DIM), &
                                     size(grid%face_exact,dim=EIGEN_DIM))
do i=1,size(grid%face_exact,dim=BASIS_DIM)
   get_face_exact_raaa(:,i,:,:) = grid%face_exact(face(1):face(2),:,:,i)
end do
end function get_face_exact_raaa

function get_elem_exact_sss(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem,basis,comp,eigen
real(my_real) :: get_elem_exact_sss
get_elem_exact_sss = grid%elem_exact(elem,comp,eigen,basis)
end function get_elem_exact_sss

function get_elem_exact_aaa(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_elem_exact_aaa(size(grid%elem_exact,dim=BASIS_DIM), &
                                    size(grid%elem_exact,dim=COMP_DIM), &
                                    size(grid%elem_exact,dim=EIGEN_DIM))
do i=1,size(grid%elem_exact,dim=BASIS_DIM)
   get_elem_exact_aaa(i,:,:) = grid%elem_exact(elem,:,:,i)
end do
end function get_elem_exact_aaa

function get_elem_exact_raaa(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem(2)
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_elem_exact_raaa(elem(2)-elem(1)+1, &
                                     size(grid%elem_exact,dim=BASIS_DIM), &
                                     size(grid%elem_exact,dim=COMP_DIM), &
                                     size(grid%elem_exact,dim=EIGEN_DIM))
do i=1,size(grid%elem_exact,dim=BASIS_DIM)
   get_elem_exact_raaa(:,i,:,:) = grid%elem_exact(elem(1):elem(2),:,:,i)
end do
end function get_elem_exact_raaa

subroutine set_vert_exact_sss(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert,comp,eigen
real(my_real), intent(in) :: val
grid%vertex_exact(vert,comp,eigen) = val
end subroutine set_vert_exact_sss

subroutine set_vert_exact_saa(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val(:,:)
grid%vertex_exact(vert,:,:) = val
end subroutine set_vert_exact_saa

subroutine set_vert_exact_aaa(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
logical, intent(in) :: vert,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
grid%vertex_exact(:,:,:) = val
end subroutine set_vert_exact_aaa

subroutine set_vert_exact_raa(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert(2)
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val(:,:,:)
grid%vertex_exact(vert(1):vert(2),:,:) = val
end subroutine set_vert_exact_raa

subroutine set_edge_exact_sss(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%edge_exact(edge,comp,eigen,basis) = val
end subroutine set_edge_exact_sss

subroutine set_edge_exact_aaa(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%edge_exact(edge,:,:,i) = val(i,:,:)
end do
end subroutine set_edge_exact_aaa

subroutine set_edge_exact_aaac(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%edge_exact(edge,:,:,:) = val
end subroutine set_edge_exact_aaac

subroutine set_edge_exact_raaa(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge(2)
logical, intent(in) :: basis, comp, eigen
real(my_real), intent(in) :: val(:,:,:,:)
integer :: i
do i=1,size(val,dim=2)
   grid%edge_exact(edge(1):edge(2),:,:,i) = val(:,i,:,:)
end do
end subroutine set_edge_exact_raaa

subroutine set_face_exact_sss(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
integer, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%face_exact(face,comp,eigen,basis) = val
end subroutine set_face_exact_sss

subroutine set_face_exact_aaa(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%face_exact(face,:,:,i) = val(i,:,:)
end do
end subroutine set_face_exact_aaa

subroutine set_face_exact_aaac(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%face_exact(face,:,:,:) = val
end subroutine set_face_exact_aaac

subroutine set_face_exact_raaa(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face(2)
logical, intent(in) :: basis, comp, eigen
real(my_real), intent(in) :: val(:,:,:,:)
integer :: i
do i=1,size(val,dim=2)
   grid%face_exact(face(1):face(2),:,:,i) = val(:,i,:,:)
end do
end subroutine set_face_exact_raaa

subroutine set_elem_exact_sss(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
integer, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%elem_exact(elem,comp,eigen,basis) = val
end subroutine set_elem_exact_sss

subroutine set_elem_exact_aaa(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%elem_exact(elem,:,:,i) = val(i,:,:)
end do
end subroutine set_elem_exact_aaa

subroutine set_elem_exact_aaac(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%elem_exact(elem,:,:,:) = val
end subroutine set_elem_exact_aaac

subroutine set_elem_exact_raaa(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem(2)
logical, intent(in) :: basis, comp, eigen
integer :: i
real(my_real), intent(in) :: val(:,:,:,:)
do i=1,size(val,dim=2)
   grid%elem_exact(elem(1):elem(2),:,:,i) = val(:,i,:,:)
end do
end subroutine set_elem_exact_raaa

subroutine copy_vert_exact_saa(from_grid,from_vert,to_grid,to_vert,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_vert, to_vert
logical, intent(in) :: comp, eigen
to_grid%vertex_exact(to_vert,:,:) = from_grid%vertex_exact(from_vert,:,:)
end subroutine copy_vert_exact_saa

subroutine copy_vert_exact_raa(from_grid,to_grid,vert,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: vert(2)
logical, intent(in) :: comp, eigen
to_grid%vertex_exact(vert(1):vert(2),:,:) = &
   from_grid%vertex_exact(vert(1):vert(2),:,:)
end subroutine copy_vert_exact_raa

subroutine copy_edge_exact_aaa(from_grid,from_edge,to_grid,to_edge,basis,comp, &
                               eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_edge, to_edge
logical, intent(in) :: basis, comp, eigen
to_grid%edge_exact(to_edge,:,:,:) = from_grid%edge_exact(from_edge,:,:,:)
end subroutine copy_edge_exact_aaa

subroutine copy_edge_exact_raaa(from_grid,to_grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: edge(2)
logical, intent(in) :: basis, comp, eigen
to_grid%edge_exact(edge(1):edge(2),:,:,:) = &
   from_grid%edge_exact(edge(1):edge(2),:,:,:)
end subroutine copy_edge_exact_raaa

subroutine copy_face_exact_aaa(from_grid,from_face,to_grid,to_face,basis,comp, &
                               eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_face, to_face
logical, intent(in) :: basis, comp, eigen
to_grid%face_exact(to_face,:,:,:) = from_grid%face_exact(from_face,:,:,:)
end subroutine copy_face_exact_aaa

subroutine copy_face_exact_raaa(from_grid,to_grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: face(2)
logical, intent(in) :: basis, comp, eigen
to_grid%face_exact(face(1):face(2),:,:,:) = &
   from_grid%face_exact(face(1):face(2),:,:,:)
end subroutine copy_face_exact_raaa

subroutine copy_elem_exact_aaa(from_grid,from_elem,to_grid,to_elem,basis,comp, &
                               eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_elem, to_elem
logical, intent(in) :: basis, comp, eigen
to_grid%elem_exact(to_elem,:,:,:) = from_grid%elem_exact(from_elem,:,:,:)
end subroutine copy_elem_exact_aaa

subroutine copy_elem_exact_raaa(from_grid,to_grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: elem(2)
logical, intent(in) :: basis, comp, eigen
to_grid%elem_exact(elem(1):elem(2),:,:,:) = &
   from_grid%elem_exact(elem(1):elem(2),:,:,:)
end subroutine copy_elem_exact_raaa

subroutine allocate_vert_exact(grid,vert,size_grid,size_comp,size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: vert, size_grid, size_comp, size_eigen
integer :: s_comp, s_eigen, astat
if (present(vert)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("vert cannot be present in allocate_vert_exact")
   stop
endif
if (.not. present(size_grid)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("size_grid must be present in allocate_vert_exact")
   stop
endif
if (present(size_comp)) then
   s_comp = size_comp
else
   s_comp = grid%system_size
endif
if (present(size_eigen)) then
   s_eigen = size_eigen
else
   s_eigen = max(1,grid%num_eval)
endif
if (associated(grid%vertex_exact)) deallocate(grid%vertex_exact)
if (size_grid==0 .or. s_comp==0 .or. s_eigen==0) then
   nullify(grid%vertex_exact)
else
   allocate(grid%vertex_exact(size_grid,s_comp,s_eigen),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in allocate_vert_exact")
      stop
   endif
endif
end subroutine allocate_vert_exact

subroutine allocate_edge_exact(grid,edge,size_grid,size_basis,size_comp, &
                               size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, astat
if (present(edge)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in allocate_edge_exact")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%edge_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (associated(grid%edge_exact)) deallocate(grid%edge_exact)
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      nullify(grid%edge_exact)
   else
      allocate(grid%edge_exact(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in allocate_edge_exact")
         stop
      endif
   endif
endif
end subroutine allocate_edge_exact

subroutine allocate_face_exact(grid,face,size_grid,size_basis,size_comp, &
                               size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, astat
if (present(face)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in allocate_face_exact")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%face_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (associated(grid%face_exact)) deallocate(grid%face_exact)
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      nullify(grid%face_exact)
   else
      allocate(grid%face_exact(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in allocate_face_exact")
         stop
      endif
   endif
endif
end subroutine allocate_face_exact

subroutine allocate_elem_exact(grid,elem,size_grid,size_basis,size_comp, &
                               size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, astat
if (present(elem)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in allocate_elem_exact")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%elem_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (associated(grid%elem_exact)) deallocate(grid%elem_exact)
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      nullify(grid%elem_exact)
   else
      allocate(grid%elem_exact(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in allocate_elem_exact")
         stop
      endif
   endif
endif
end subroutine allocate_elem_exact

subroutine reallocate_vert_exact(grid,vert,size_grid,size_comp,size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: vert, size_grid, size_comp, size_eigen
integer :: s_comp, s_eigen, ncopy, ncopyss, ncopyev, astat
real(my_real), pointer :: temp(:,:,:)
if (present(vert)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("vert must not be present in reallocate_vert_exact")
   stop
endif
if (.not. present(size_grid)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("size_grid must be present in reallocate_vert_exact")
   stop
endif
if (present(size_comp)) then
   s_comp = size_comp
else
   s_comp = grid%system_size
endif
if (present(size_eigen)) then
   s_eigen = size_eigen
else
   s_eigen = max(1,grid%num_eval)
endif
if (size_grid==0 .or. s_comp==0 .or. s_eigen==0) then
   if (associated(grid%vertex_exact)) deallocate(grid%vertex_exact)
   nullify(grid%vertex_exact)
else
   allocate(temp(size_grid,s_comp,s_eigen),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in reallocate_vert_exact")
      stop
   endif
   temp(1:min(size_grid,grid%biggest_vert),:,:) = 0.0_my_real
   if (associated(grid%vertex_exact)) then
      ncopy = min(size_grid,size_vert_exact(grid,dim=GRID_DIM))
      ncopyss = min(s_comp,size_vert_exact(grid,dim=COMP_DIM))
      ncopyev = min(s_eigen,size_vert_exact(grid,dim=EIGEN_DIM))
      if (ncopy > 0) then
         temp(1:ncopy,1:ncopyss,1:ncopyev) = &
            grid%vertex_exact(1:ncopy,1:ncopyss,1:ncopyev)
      endif
      deallocate(grid%vertex_exact)
   endif
   grid%vertex_exact => temp
   nullify(temp)
endif
end subroutine reallocate_vert_exact

subroutine reallocate_edge_exact(grid,edge,size_grid,size_basis,size_comp, &
                                size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, ncopy, ncopyss, ncopyev, ncopyb, astat
real(my_real), pointer :: temp(:,:,:,:)
if (present(edge)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in reallocate_edge_exact")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%edge_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      if (associated(grid%edge_exact)) deallocate(grid%edge_exact)
      nullify(grid%edge_exact)
   else
      allocate(temp(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in reallocate_edge_exact")
         stop
      endif
      temp(1:min(size_grid,grid%biggest_edge),:,:,:) = 0.0_my_real
      if (associated(grid%edge_exact)) then
         ncopy = min(size_grid,size_edge_exact(grid,dim=GRID_DIM))
         ncopyb  = min(s_basis,size_edge_exact(grid,dim=BASIS_DIM))
         ncopyss = min(s_comp,size_edge_exact(grid,dim=COMP_DIM))
         ncopyev = min(s_eigen,size_edge_exact(grid,dim=EIGEN_DIM))
         if (ncopy > 0 .and. ncopyb > 0) then
            temp(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb) = &
               grid%edge_exact(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb)
         endif
         deallocate(grid%edge_exact)
      endif
      grid%edge_exact => temp
      nullify(temp)
   endif
endif
end subroutine reallocate_edge_exact

subroutine reallocate_face_exact(grid,face,size_grid,size_basis,size_comp, &
                                 size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, ncopy, ncopyss, ncopyev, ncopyb, astat
real(my_real), pointer :: temp(:,:,:,:)
if (present(face)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in reallocate_face_exact")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%face_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      if (associated(grid%face_exact)) deallocate(grid%face_exact)
      nullify(grid%face_exact)
   else
      allocate(temp(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in reallocate_face_exact")
         stop
      endif
      temp(1:min(size_grid,grid%biggest_face),:,:,:) = 0.0_my_real
      if (associated(grid%face_exact)) then
         ncopy = min(size_grid,size_face_exact(grid,dim=GRID_DIM))
         ncopyb  = min(s_basis,size_face_exact(grid,dim=BASIS_DIM))
         ncopyss = min(s_comp,size_face_exact(grid,dim=COMP_DIM))
         ncopyev = min(s_eigen,size_face_exact(grid,dim=EIGEN_DIM))
         if (ncopy > 0 .and. ncopyb > 0) then
            temp(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb) = &
               grid%face_exact(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb)
         endif
         deallocate(grid%face_exact)
      endif
      grid%face_exact => temp
      nullify(temp)
   endif
endif
end subroutine reallocate_face_exact

subroutine reallocate_elem_exact(grid,elem,size_grid,size_basis,size_comp, &
                                size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, ncopy, ncopyss, ncopyev, ncopyb, astat
real(my_real), pointer :: temp(:,:,:,:)
if (present(elem)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in reallocate_elem_exact")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%elem_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      if (associated(grid%elem_exact)) deallocate(grid%elem_exact)
      nullify(grid%elem_exact)
   else
      allocate(temp(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in reallocate_elem_exact")
         stop
      endif
      temp(1:min(size_grid,grid%biggest_elem),:,:,:) = 0.0_my_real
      if (associated(grid%elem_exact)) then
         ncopy = min(size_grid,size_elem_exact(grid,dim=GRID_DIM))
         ncopyb  = min(s_basis,size_elem_exact(grid,dim=BASIS_DIM))
         ncopyss = min(s_comp,size_elem_exact(grid,dim=COMP_DIM))
         ncopyev = min(s_eigen,size_elem_exact(grid,dim=EIGEN_DIM))
         if (ncopy > 0 .and. ncopyb > 0) then
            temp(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb) = &
               grid%elem_exact(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb)
         endif
         deallocate(grid%elem_exact)
      endif
      grid%elem_exact => temp
      nullify(temp)
   endif
endif
end subroutine reallocate_elem_exact

subroutine deallocate_vert_exact(grid)
type(grid_type), intent(inout) :: grid
deallocate(grid%vertex_exact)
end subroutine deallocate_vert_exact

subroutine deallocate_edge_exact(grid,edge)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge
if (present(edge)) then
! do nothing
else
   deallocate(grid%edge_exact)
endif
end subroutine deallocate_edge_exact

subroutine deallocate_face_exact(grid,face)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face
if (present(face)) then
! do nothing
else
   deallocate(grid%face_exact)
endif
end subroutine deallocate_face_exact

subroutine deallocate_elem_exact(grid,elem)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem
if (present(elem)) then
! do nothing
else
   deallocate(grid%elem_exact)
endif
end subroutine deallocate_elem_exact

function associated_vert_exact(grid)
type(grid_type), intent(in) :: grid
logical :: associated_vert_exact
associated_vert_exact = associated(grid%vertex_exact)
end function associated_vert_exact

function associated_edge_exact(grid,edge)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: edge
logical :: associated_edge_exact
if (present(edge)) then
   if (associated(grid%edge_exact)) then
      associated_edge_exact = edge <= size(grid%edge_exact,dim=GRID_DIM)
   else
      associated_edge_exact = .false.
   endif
else
   associated_edge_exact = associated(grid%edge_exact)
endif
end function associated_edge_exact

function associated_face_exact(grid,face)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: face
logical :: associated_face_exact
if (present(face)) then
   if (associated(grid%face_exact)) then
      associated_face_exact = face <= size(grid%face_exact,dim=GRID_DIM)
   else
      associated_face_exact = .false.
   endif
else
   associated_face_exact = associated(grid%face_exact)
endif
end function associated_face_exact

function associated_elem_exact(grid,elem)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: elem
logical :: associated_elem_exact
if (present(elem)) then
   if (associated(grid%elem_exact)) then
      associated_elem_exact = elem <= size(grid%elem_exact,dim=GRID_DIM)
   else
      associated_elem_exact = .false.
   endif
else
   associated_elem_exact = associated(grid%elem_exact)
endif
end function associated_elem_exact

subroutine nullify_vert_exact(grid)
type(grid_type), intent(inout) :: grid
nullify(grid%vertex_exact)
end subroutine nullify_vert_exact

subroutine nullify_edge_exact(grid,edge)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge
if (present(edge)) then
! do nothing
else
   nullify(grid%edge_exact)
endif
end subroutine nullify_edge_exact

subroutine nullify_face_exact(grid,face)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face
if (present(face)) then
! do nothing
else
   nullify(grid%face_exact)
endif
end subroutine nullify_face_exact

subroutine nullify_elem_exact(grid,elem)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem
if (present(elem)) then
! do nothing
else
   nullify(grid%elem_exact)
endif
end subroutine nullify_elem_exact

function size_vert_exact(grid,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: dim
integer :: size_vert_exact
integer :: copydim
if (associated_vert_exact(grid)) then
   if (present(dim)) then
      copydim = dim
      size_vert_exact = size(grid%vertex_exact,dim=copydim)
   else
      size_vert_exact = size(grid%vertex_exact)
   endif
else
   size_vert_exact = 0
endif
end function size_vert_exact

function size_edge_exact(grid,edge,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: edge
integer, intent(in), optional :: dim
integer :: size_edge_exact
integer :: copydim
if (associated_edge_exact(grid,edge)) then
   if (present(edge)) then
      if (present(dim)) then
         copydim = dim
         size_edge_exact = size(grid%edge_exact,dim=copydim)
      else
         size_edge_exact = size(grid%edge_exact,dim=BASIS_DIM) * &
                           size(grid%edge_exact,dim=COMP_DIM) * &
                           size(grid%edge_exact,dim=EIGEN_DIM)
      endif
   else
      if (present(dim)) then
         copydim = dim
         size_edge_exact = size(grid%edge_exact,dim=copydim)
      else
         size_edge_exact = size(grid%edge_exact)
      endif
   endif
else
   size_edge_exact = 0
endif
end function size_edge_exact

function size_face_exact(grid,face,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: face
integer, intent(in), optional :: dim
integer :: size_face_exact
integer :: copydim
if (associated_face_exact(grid,face)) then
   if (present(face)) then
      if (present(dim)) then
         copydim = dim
         size_face_exact = size(grid%face_exact,dim=copydim)
      else
         size_face_exact = size(grid%face_exact,dim=BASIS_DIM) * &
                           size(grid%face_exact,dim=COMP_DIM) * &
                           size(grid%face_exact,dim=EIGEN_DIM)
      endif
   else
      if (present(dim)) then
         copydim = dim
         size_face_exact = size(grid%face_exact,dim=copydim)
      else
         size_face_exact = size(grid%face_exact)
      endif
   endif
else
   size_face_exact = 0
endif
end function size_face_exact

function size_elem_exact(grid,elem,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: elem
integer, intent(in), optional :: dim
integer :: size_elem_exact
integer :: copydim
if (associated_elem_exact(grid,elem)) then
   if (present(elem)) then
      if (present(dim)) then
         copydim = dim
         size_elem_exact = size(grid%elem_exact,dim=copydim)
      else
         size_elem_exact = size(grid%elem_exact,dim=BASIS_DIM) * &
                           size(grid%elem_exact,dim=COMP_DIM) * &
                           size(grid%elem_exact,dim=EIGEN_DIM)
      endif
   else
      if (present(dim)) then
         copydim = dim
         size_elem_exact = size(grid%elem_exact,dim=copydim)
      else
         size_elem_exact = size(grid%elem_exact)
      endif
   endif
else
   size_elem_exact = 0
endif
end function size_elem_exact

function get_vert_oldsoln_sss(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert,comp,eigen
real(my_real) :: get_vert_oldsoln_sss
get_vert_oldsoln_sss = grid%vertex_oldsoln(vert,comp,eigen)
end function get_vert_oldsoln_sss

function get_vert_oldsoln_saa(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert
logical, intent(in) :: comp,eigen
real(my_real) :: get_vert_oldsoln_saa(size(grid%vertex_oldsoln,dim=COMP_DIM), &
                                      size(grid%vertex_oldsoln,dim=EIGEN_DIM))
get_vert_oldsoln_saa = grid%vertex_oldsoln(vert,:,:)
end function get_vert_oldsoln_saa

function get_vert_oldsoln_aaa(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
logical, intent(in) :: vert,comp,eigen
real(my_real) :: get_vert_oldsoln_aaa(size(grid%vertex_oldsoln,dim=GRID_DIM), &
                                      size(grid%vertex_oldsoln,dim=COMP_DIM), &
                                      size(grid%vertex_oldsoln,dim=EIGEN_DIM))
get_vert_oldsoln_aaa = grid%vertex_oldsoln(:,:,:)
end function get_vert_oldsoln_aaa

function get_vert_oldsoln_raa(grid,vert,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: vert(2)
logical, intent(in) :: comp,eigen
real(my_real) :: get_vert_oldsoln_raa(vert(2)-vert(1)+1, &
                                      size(grid%vertex_oldsoln,dim=COMP_DIM), &
                                      size(grid%vertex_oldsoln,dim=EIGEN_DIM))
get_vert_oldsoln_raa = grid%vertex_oldsoln(vert(1):vert(2),:,:)
end function get_vert_oldsoln_raa

function get_edge_oldsoln_sss(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge,basis,comp,eigen
real(my_real) :: get_edge_oldsoln_sss
get_edge_oldsoln_sss = grid%edge_oldsoln(edge,comp,eigen,basis)
end function get_edge_oldsoln_sss

function get_edge_oldsoln_saa(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge,basis
logical, intent(in) :: comp,eigen
real(my_real) :: get_edge_oldsoln_saa(size(grid%edge_oldsoln,dim=COMP_DIM), &
                                      size(grid%edge_oldsoln,dim=EIGEN_DIM))
get_edge_oldsoln_saa = grid%edge_oldsoln(edge,:,:,basis)
end function get_edge_oldsoln_saa

function get_edge_oldsoln_aaa(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_edge_oldsoln_aaa(size(grid%edge_oldsoln,dim=BASIS_DIM), &
                                      size(grid%edge_oldsoln,dim=COMP_DIM), &
                                      size(grid%edge_oldsoln,dim=EIGEN_DIM))
do i=1,size(grid%edge_oldsoln,dim=BASIS_DIM)
   get_edge_oldsoln_aaa(i,:,:) = grid%edge_oldsoln(edge,:,:,i)
end do
end function get_edge_oldsoln_aaa

function get_edge_oldsoln_raaa(grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: edge(2)
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_edge_oldsoln_raaa(edge(2)-edge(1)+1, &
                                       size(grid%edge_oldsoln,dim=BASIS_DIM), &
                                       size(grid%edge_oldsoln,dim=COMP_DIM), &
                                       size(grid%edge_oldsoln,dim=EIGEN_DIM))
do i=1,size(grid%edge_oldsoln,dim=BASIS_DIM)
   get_edge_oldsoln_raaa(:,i,:,:) = grid%edge_oldsoln(edge(1):edge(2),:,:,i)
end do
end function get_edge_oldsoln_raaa

function get_face_oldsoln_sss(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face,basis,comp,eigen
real(my_real) :: get_face_oldsoln_sss
get_face_oldsoln_sss = grid%face_oldsoln(face,comp,eigen,basis)
end function get_face_oldsoln_sss

function get_face_oldsoln_aaa(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_face_oldsoln_aaa(size(grid%face_oldsoln,dim=BASIS_DIM), &
                                      size(grid%face_oldsoln,dim=COMP_DIM), &
                                      size(grid%face_oldsoln,dim=EIGEN_DIM))
do i=1,size(grid%face_oldsoln,dim=BASIS_DIM)
   get_face_oldsoln_aaa(i,:,:) = grid%face_oldsoln(face,:,:,i)
end do
end function get_face_oldsoln_aaa

function get_face_oldsoln_raaa(grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: face(2)
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_face_oldsoln_raaa(face(2)-face(1)+1, &
                                       size(grid%face_oldsoln,dim=BASIS_DIM), &
                                       size(grid%face_oldsoln,dim=COMP_DIM), &
                                       size(grid%face_oldsoln,dim=EIGEN_DIM))
do i=1,size(grid%face_oldsoln,dim=BASIS_DIM)
   get_face_oldsoln_raaa(:,i,:,:) = grid%face_oldsoln(face(1):face(2),:,:,i)
end do
end function get_face_oldsoln_raaa

function get_elem_oldsoln_sss(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem,basis,comp,eigen
real(my_real) :: get_elem_oldsoln_sss
get_elem_oldsoln_sss = grid%elem_oldsoln(elem,comp,eigen,basis)
end function get_elem_oldsoln_sss

function get_elem_oldsoln_aaa(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_elem_oldsoln_aaa(size(grid%elem_oldsoln,dim=BASIS_DIM), &
                                      size(grid%elem_oldsoln,dim=COMP_DIM), &
                                      size(grid%elem_oldsoln,dim=EIGEN_DIM))
do i=1,size(grid%elem_oldsoln,dim=BASIS_DIM)
   get_elem_oldsoln_aaa(i,:,:) = grid%elem_oldsoln(elem,:,:,i)
end do
end function get_elem_oldsoln_aaa

function get_elem_oldsoln_raaa(grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: grid
integer, intent(in) :: elem(2)
logical, intent(in) :: basis,comp,eigen
integer :: i
real(my_real) :: get_elem_oldsoln_raaa(elem(2)-elem(1)+1, &
                                       size(grid%elem_oldsoln,dim=BASIS_DIM), &
                                       size(grid%elem_oldsoln,dim=COMP_DIM), &
                                       size(grid%elem_oldsoln,dim=EIGEN_DIM))
do i=1,size(grid%elem_oldsoln,dim=BASIS_DIM)
   get_elem_oldsoln_raaa(:,i,:,:) = grid%elem_oldsoln(elem(1):elem(2),:,:,i)
end do
end function get_elem_oldsoln_raaa

subroutine set_vert_oldsoln_sss(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert,comp,eigen
real(my_real), intent(in) :: val
grid%vertex_oldsoln(vert,comp,eigen) = val
end subroutine set_vert_oldsoln_sss

subroutine set_vert_oldsoln_saa(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val(:,:)
grid%vertex_oldsoln(vert,:,:) = val
end subroutine set_vert_oldsoln_saa

subroutine set_vert_oldsoln_aaa(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
logical, intent(in) :: vert,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
grid%vertex_oldsoln = val
end subroutine set_vert_oldsoln_aaa

subroutine set_vert_oldsoln_raa(grid,vert,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: vert(2)
logical, intent(in) :: comp,eigen
real(my_real), intent(in) :: val(:,:,:)
grid%vertex_oldsoln(vert(1):vert(2),:,:) = val
end subroutine set_vert_oldsoln_raa

subroutine set_edge_oldsoln_sss(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
integer, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%edge_oldsoln(edge,comp,eigen,basis) = val
end subroutine set_edge_oldsoln_sss

subroutine set_edge_oldsoln_aaa(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%edge_oldsoln(edge,:,:,i) = val(i,:,:)
end do
end subroutine set_edge_oldsoln_aaa

subroutine set_edge_oldsoln_aaac(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%edge_oldsoln(edge,:,:,:) = val
end subroutine set_edge_oldsoln_aaac

subroutine set_edge_oldsoln_raaa(grid,edge,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: edge(2)
logical, intent(in) :: basis, comp, eigen
real(my_real), intent(in) :: val(:,:,:,:)
integer :: i
do i=1,size(val,dim=2)
   grid%edge_oldsoln(edge(1):edge(2),:,:,i) = val(:,i,:,:)
end do
end subroutine set_edge_oldsoln_raaa

subroutine set_face_oldsoln_sss(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
integer, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%face_oldsoln(face,comp,eigen,basis) = val
end subroutine set_face_oldsoln_sss

subroutine set_face_oldsoln_aaa(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%face_oldsoln(face,:,:,i) = val(i,:,:)
end do
end subroutine set_face_oldsoln_aaa

subroutine set_face_oldsoln_raaa(grid,face,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: face(2)
logical, intent(in) :: basis, comp, eigen
real(my_real), intent(in) :: val(:,:,:,:)
integer :: i
do i=1,size(val,dim=2)
   grid%face_oldsoln(face(1):face(2),:,:,i) = val(:,i,:,:)
end do
end subroutine set_face_oldsoln_raaa

subroutine set_elem_oldsoln_sss(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
integer, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%elem_oldsoln(elem,comp,eigen,basis) = val
end subroutine set_elem_oldsoln_sss

subroutine set_elem_oldsoln_aaa(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val(:,:,:)
integer :: i
do i=1,size(val,dim=1)
   grid%elem_oldsoln(elem,:,:,i) = val(i,:,:)
end do
end subroutine set_elem_oldsoln_aaa

subroutine set_elem_oldsoln_aaac(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
logical, intent(in) :: basis,comp,eigen
real(my_real), intent(in) :: val
grid%elem_oldsoln(elem,:,:,:) = val
end subroutine set_elem_oldsoln_aaac

subroutine set_elem_oldsoln_raaa(grid,elem,basis,comp,eigen,val)
type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem(2)
logical, intent(in) :: basis, comp, eigen
real(my_real), intent(in) :: val(:,:,:,:)
integer :: i
do i=1,size(val,dim=2)
   grid%elem_oldsoln(elem(1):elem(2),:,:,i) = val(:,i,:,:)
end do
end subroutine set_elem_oldsoln_raaa

subroutine copy_vert_oldsoln_saa(from_grid,from_vert,to_grid,to_vert,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_vert, to_vert
logical, intent(in) :: comp, eigen
to_grid%vertex_oldsoln(to_vert,:,:) = from_grid%vertex_oldsoln(from_vert,:,:)
end subroutine copy_vert_oldsoln_saa

subroutine copy_vert_oldsoln_raa(from_grid,to_grid,vert,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: vert(2)
logical, intent(in) :: comp, eigen
to_grid%vertex_oldsoln(vert(1):vert(2),:,:) = &
   from_grid%vertex_oldsoln(vert(1):vert(2),:,:)
end subroutine copy_vert_oldsoln_raa

subroutine copy_edge_oldsoln_aaa(from_grid,from_edge,to_grid,to_edge,basis, &
                                 comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_edge, to_edge
logical, intent(in) :: basis, comp, eigen
to_grid%edge_oldsoln(to_edge,:,:,:) = from_grid%edge_oldsoln(from_edge,:,:,:)
end subroutine copy_edge_oldsoln_aaa

subroutine copy_edge_oldsoln_raaa(from_grid,to_grid,edge,basis,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: edge(2)
logical, intent(in) :: basis, comp, eigen
to_grid%edge_oldsoln(edge(1):edge(2),:,:,:) = &
   from_grid%edge_oldsoln(edge(1):edge(2),:,:,:)
end subroutine copy_edge_oldsoln_raaa

subroutine copy_face_oldsoln_aaa(from_grid,from_face,to_grid,to_face,basis, &
                                 comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_face, to_face
logical, intent(in) :: basis, comp, eigen
to_grid%face_oldsoln(to_face,:,:,:) = from_grid%face_oldsoln(from_face,:,:,:)
end subroutine copy_face_oldsoln_aaa

subroutine copy_face_oldsoln_raaa(from_grid,to_grid,face,basis,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: face(2)
logical, intent(in) :: basis, comp, eigen
to_grid%face_oldsoln(face(1):face(2),:,:,:) = &
   from_grid%face_oldsoln(face(1):face(2),:,:,:)
end subroutine copy_face_oldsoln_raaa

subroutine copy_elem_oldsoln_aaa(from_grid,from_elem,to_grid,to_elem,basis, &
                                 comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: from_elem, to_elem
logical, intent(in) :: basis, comp, eigen
to_grid%elem_oldsoln(to_elem,:,:,:) = from_grid%elem_oldsoln(from_elem,:,:,:)
end subroutine copy_elem_oldsoln_aaa

subroutine copy_elem_oldsoln_raaa(from_grid,to_grid,elem,basis,comp,eigen)
type(grid_type), intent(in) :: from_grid
type(grid_type), intent(inout) :: to_grid
integer, intent(in) :: elem(2)
logical, intent(in) :: basis, comp, eigen
to_grid%elem_oldsoln(elem(1):elem(2),:,:,:) = &
   from_grid%elem_oldsoln(elem(1):elem(2),:,:,:)
end subroutine copy_elem_oldsoln_raaa

subroutine allocate_vert_oldsoln(grid,vert,size_grid,size_comp,size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: vert, size_grid, size_comp, size_eigen
integer :: s_comp, s_eigen, astat
if (present(vert)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("vert cannot be present in allocate_vert_oldsoln")
   stop
endif
if (.not. present(size_grid)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("size_grid must be present in allocate_vert_oldsoln")
   stop
endif
if (present(size_comp)) then
   s_comp = size_comp
else
   s_comp = grid%system_size
endif
if (present(size_eigen)) then
   s_eigen = size_eigen
else
   s_eigen = max(1,grid%num_eval)
endif
if (associated(grid%vertex_oldsoln)) deallocate(grid%vertex_oldsoln)
if (size_grid==0 .or. s_comp==0 .or. s_eigen==0) then
   nullify(grid%vertex_oldsoln)
else
   allocate(grid%vertex_oldsoln(size_grid,s_comp,s_eigen),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in allocate_vert_oldsoln")
      stop
   endif
endif
end subroutine allocate_vert_oldsoln

subroutine allocate_edge_oldsoln(grid,edge,size_grid,size_basis,size_comp, &
                                 size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, astat
if (present(edge)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in allocate_edge_oldsoln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%edge_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (associated(grid%edge_oldsoln)) deallocate(grid%edge_oldsoln)
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      nullify(grid%edge_oldsoln)
   else
      allocate(grid%edge_oldsoln(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in allocate_edge_oldsoln")
         stop
      endif
   endif
endif
end subroutine allocate_edge_oldsoln

subroutine allocate_face_oldsoln(grid,face,size_grid,size_basis,size_comp, &
                                 size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, astat
if (present(face)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in allocate_face_oldsoln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%face_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (associated(grid%face_oldsoln)) deallocate(grid%face_oldsoln)
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      nullify(grid%face_oldsoln)
   else
      allocate(grid%face_oldsoln(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in allocate_face_oldsoln")
         stop
      endif
   endif
endif
end subroutine allocate_face_oldsoln

subroutine allocate_elem_oldsoln(grid,elem,size_grid,size_basis,size_comp, &
                                 size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, astat
if (present(elem)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in allocate_elem_oldsoln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%elem_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (associated(grid%elem_oldsoln)) deallocate(grid%elem_oldsoln)
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      nullify(grid%elem_oldsoln)
   else
      allocate(grid%elem_oldsoln(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in allocate_elem_oldsoln")
         stop
      endif
   endif
endif
end subroutine allocate_elem_oldsoln

subroutine reallocate_vert_oldsoln(grid,vert,size_grid,size_comp,size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: vert, size_grid, size_comp, size_eigen
integer :: s_comp, s_eigen, ncopy, ncopyss, ncopyev, astat
real(my_real), pointer :: temp(:,:,:)
if (present(vert)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("vert must not be present in reallocate_vert_oldsoln")
   stop
endif
if (.not. present(size_grid)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("size_grid must be present in reallocate_vert_oldsoln")
   stop
endif
if (present(size_comp)) then
   s_comp = size_comp
else
   s_comp = grid%system_size
endif
if (present(size_eigen)) then
   s_eigen = size_eigen
else
   s_eigen = max(1,grid%num_eval)
endif
if (size_grid==0 .or. s_comp==0 .or. s_eigen==0) then
   if (associated(grid%vertex_oldsoln)) deallocate(grid%vertex_oldsoln)
   nullify(grid%vertex_oldsoln)
else
   allocate(temp(size_grid,s_comp,s_eigen),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in reallocate_vert_oldsoln")
      stop
   endif
   temp(1:min(size_grid,grid%biggest_vert),:,:) = 0.0_my_real
   if (associated(grid%vertex_oldsoln)) then
      ncopy = min(size_grid,size_vert_oldsoln(grid,dim=GRID_DIM))
      ncopyss = min(s_comp,size_vert_oldsoln(grid,dim=COMP_DIM))
      ncopyev = min(s_eigen,size_vert_oldsoln(grid,dim=EIGEN_DIM))
      if (ncopy > 0) then
         temp(1:ncopy,1:ncopyss,1:ncopyev) = &
            grid%vertex_oldsoln(1:ncopy,1:ncopyss,1:ncopyev)
      endif
      deallocate(grid%vertex_oldsoln)
   endif
   grid%vertex_oldsoln => temp
   nullify(temp)
endif
end subroutine reallocate_vert_oldsoln

subroutine reallocate_edge_oldsoln(grid,edge,size_grid,size_basis,size_comp, &
                                   size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, ncopy, ncopyss, ncopyev, ncopyb, astat
real(my_real), pointer :: temp(:,:,:,:)
if (present(edge)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in reallocate_edge_oldsoln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%edge_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      if (associated(grid%edge_oldsoln)) deallocate(grid%edge_oldsoln)
      nullify(grid%edge_oldsoln)
   else
      allocate(temp(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in reallocate_edge_oldsoln")
         stop
      endif
      temp(1:min(size_grid,grid%biggest_edge),:,:,:) = 0.0_my_real
      if (associated(grid%edge_oldsoln)) then
         ncopy = min(size_grid,size_edge_oldsoln(grid,dim=GRID_DIM))
         ncopyb  = min(s_basis,size_edge_oldsoln(grid,dim=BASIS_DIM))
         ncopyss = min(s_comp,size_edge_oldsoln(grid,dim=COMP_DIM))
         ncopyev = min(s_eigen,size_edge_oldsoln(grid,dim=EIGEN_DIM))
         if (ncopy > 0 .and. ncopyb > 0) then
            temp(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb) = &
               grid%edge_oldsoln(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb)
         endif
         deallocate(grid%edge_oldsoln)
      endif
      grid%edge_oldsoln => temp
      nullify(temp)
   endif
endif
end subroutine reallocate_edge_oldsoln

subroutine reallocate_face_oldsoln(grid,face,size_grid,size_basis,size_comp, &
                                   size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, ncopy, ncopyss, ncopyev, ncopyb, astat
real(my_real), pointer :: temp(:,:,:,:)
if (present(face)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in reallocate_face_oldsoln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%face_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      if (associated(grid%face_oldsoln)) deallocate(grid%face_oldsoln)
      nullify(grid%face_oldsoln)
   else
      allocate(temp(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in reallocate_face_oldsoln")
         stop
      endif
      temp(1:min(size_grid,grid%biggest_face),:,:,:) = 0.0_my_real
      if (associated(grid%face_oldsoln)) then
         ncopy = min(size_grid,size_face_oldsoln(grid,dim=GRID_DIM))
         ncopyb  = min(s_basis,size_face_oldsoln(grid,dim=BASIS_DIM))
         ncopyss = min(s_comp,size_face_oldsoln(grid,dim=COMP_DIM))
         ncopyev = min(s_eigen,size_face_oldsoln(grid,dim=EIGEN_DIM))
         if (ncopy > 0 .and. ncopyb > 0) then
            temp(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb) = &
               grid%face_oldsoln(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb)
         endif
         deallocate(grid%face_oldsoln)
      endif
      grid%face_oldsoln => temp
      nullify(temp)
   endif
endif
end subroutine reallocate_face_oldsoln

subroutine reallocate_elem_oldsoln(grid,elem,size_grid,size_basis,size_comp, &
                                   size_eigen)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem, size_grid, size_basis, size_comp, &
                                 size_eigen
integer :: s_basis, s_comp, s_eigen, ncopy, ncopyss, ncopyev, ncopyb, astat
real(my_real), pointer :: temp(:,:,:,:)
if (present(elem)) then
! do nothing
else
   if (.not. present(size_grid)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("size_grid must be present in reallocate_elem_oldsoln")
      stop
   endif
   if (present(size_basis)) then
      s_basis = size_basis
   else
      s_basis = grid%elem_dim_basis
   endif
   if (present(size_comp)) then
      s_comp = size_comp
   else
      s_comp = grid%system_size
   endif
   if (present(size_eigen)) then
      s_eigen = size_eigen
   else
      s_eigen = max(1,grid%num_eval)
   endif
   if (size_grid==0 .or. s_basis==0 .or. s_comp==0 .or. s_eigen==0) then
      if (associated(grid%elem_oldsoln)) deallocate(grid%elem_oldsoln)
      nullify(grid%elem_oldsoln)
   else
      allocate(temp(size_grid,s_comp,s_eigen,s_basis),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in reallocate_elem_oldsoln")
         stop
      endif
      temp(1:min(size_grid,grid%biggest_elem),:,:,:) = 0.0_my_real
      if (associated(grid%elem_oldsoln)) then
         ncopy = min(size_grid,size_elem_oldsoln(grid,dim=GRID_DIM))
         ncopyb  = min(s_basis,size_elem_oldsoln(grid,dim=BASIS_DIM))
         ncopyss = min(s_comp,size_elem_oldsoln(grid,dim=COMP_DIM))
         ncopyev = min(s_eigen,size_elem_oldsoln(grid,dim=EIGEN_DIM))
         if (ncopy > 0 .and. ncopyb > 0) then
            temp(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb) = &
               grid%elem_oldsoln(1:ncopy,1:ncopyss,1:ncopyev,1:ncopyb)
         endif
         deallocate(grid%elem_oldsoln)
      endif
      grid%elem_oldsoln => temp
      nullify(temp)
   endif
endif
end subroutine reallocate_elem_oldsoln

subroutine deallocate_vert_oldsoln(grid)
type(grid_type), intent(inout) :: grid
deallocate(grid%vertex_oldsoln)
end subroutine deallocate_vert_oldsoln

subroutine deallocate_edge_oldsoln(grid,edge)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge
if (present(edge)) then
! do nothing
else
   deallocate(grid%edge_oldsoln)
endif
end subroutine deallocate_edge_oldsoln

subroutine deallocate_face_oldsoln(grid,face)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face
if (present(face)) then
! do nothing
else
   deallocate(grid%face_oldsoln)
endif
end subroutine deallocate_face_oldsoln

subroutine deallocate_elem_oldsoln(grid,elem)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem
if (present(elem)) then
! do nothing
else
   deallocate(grid%elem_oldsoln)
endif
end subroutine deallocate_elem_oldsoln

function associated_vert_oldsoln(grid)
type(grid_type), intent(in) :: grid
logical :: associated_vert_oldsoln
associated_vert_oldsoln = associated(grid%vertex_oldsoln)
end function associated_vert_oldsoln

function associated_edge_oldsoln(grid,edge)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: edge
logical :: associated_edge_oldsoln
if (present(edge)) then
   if (associated(grid%edge_oldsoln)) then
      associated_edge_oldsoln = edge <= size(grid%edge_oldsoln,dim=GRID_DIM)
   else
      associated_edge_oldsoln = .false.
   endif
else
   associated_edge_oldsoln = associated(grid%edge_oldsoln)
endif
end function associated_edge_oldsoln

function associated_face_oldsoln(grid,face)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: face
logical :: associated_face_oldsoln
if (present(face)) then
   if (associated(grid%face_oldsoln)) then
      associated_face_oldsoln = face <= size(grid%face_oldsoln,dim=GRID_DIM)
   else
      associated_face_oldsoln = .false.
   endif
else
   associated_face_oldsoln = associated(grid%face_oldsoln)
endif
end function associated_face_oldsoln

function associated_elem_oldsoln(grid,elem)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: elem
logical :: associated_elem_oldsoln
if (present(elem)) then
   if (associated(grid%elem_oldsoln)) then
      associated_elem_oldsoln = elem <= size(grid%elem_oldsoln,dim=GRID_DIM)
   else
      associated_elem_oldsoln = .false.
   endif
else
   associated_elem_oldsoln = associated(grid%elem_oldsoln)
endif
end function associated_elem_oldsoln

subroutine nullify_vert_oldsoln(grid)
type(grid_type), intent(inout) :: grid
nullify(grid%vertex_oldsoln)
end subroutine nullify_vert_oldsoln

subroutine nullify_edge_oldsoln(grid,edge)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: edge
if (present(edge)) then
! do nothing
else
   nullify(grid%edge_oldsoln)
endif
end subroutine nullify_edge_oldsoln

subroutine nullify_face_oldsoln(grid,face)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: face
if (present(face)) then
! do nothing
else
   nullify(grid%face_oldsoln)
endif
end subroutine nullify_face_oldsoln

subroutine nullify_elem_oldsoln(grid,elem)
type(grid_type), intent(inout) :: grid
integer, intent(in), optional :: elem
if (present(elem)) then
! do nothing
else
   nullify(grid%elem_oldsoln)
endif
end subroutine nullify_elem_oldsoln

function size_vert_oldsoln(grid,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: dim
integer :: size_vert_oldsoln
integer :: copydim
if (associated_vert_oldsoln(grid)) then
   if (present(dim)) then
      copydim = dim
      size_vert_oldsoln = size(grid%vertex_oldsoln,dim=copydim)
   else
      size_vert_oldsoln = size(grid%vertex_oldsoln)
   endif
else
   size_vert_oldsoln = 0
endif
end function size_vert_oldsoln

function size_edge_oldsoln(grid,edge,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: edge
integer, intent(in), optional :: dim
integer :: size_edge_oldsoln
integer :: copydim
if (associated_edge_oldsoln(grid,edge)) then
   if (present(edge)) then
      if (present(dim)) then
         copydim = dim
         size_edge_oldsoln = size(grid%edge_oldsoln,dim=copydim)
      else
         size_edge_oldsoln = size(grid%edge_oldsoln,dim=BASIS_DIM) * &
                             size(grid%edge_oldsoln,dim=COMP_DIM) * &
                             size(grid%edge_oldsoln,dim=EIGEN_DIM)
      endif
   else
      if (present(dim)) then
         copydim = dim
         size_edge_oldsoln = size(grid%edge_oldsoln,dim=copydim)
      else
         size_edge_oldsoln = size(grid%edge_oldsoln)
      endif
   endif
else
   size_edge_oldsoln = 0
endif
end function size_edge_oldsoln

function size_face_oldsoln(grid,face,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: face
integer, intent(in), optional :: dim
integer :: size_face_oldsoln
integer :: copydim
if (associated_face_oldsoln(grid,face)) then
   if (present(face)) then
      if (present(dim)) then
         copydim = dim
         size_face_oldsoln = size(grid%face_oldsoln,dim=copydim)
      else
         size_face_oldsoln = size(grid%face_oldsoln,dim=BASIS_DIM) * &
                             size(grid%face_oldsoln,dim=COMP_DIM) * &
                             size(grid%face_oldsoln,dim=EIGEN_DIM)
      endif
   else
      if (present(dim)) then
         copydim = dim
         size_face_oldsoln = size(grid%face_oldsoln,dim=copydim)
      else
         size_face_oldsoln = size(grid%face_oldsoln)
      endif
   endif
else
   size_face_oldsoln = 0
endif
end function size_face_oldsoln

function size_elem_oldsoln(grid,elem,dim)
type(grid_type), intent(in) :: grid
integer, intent(in), optional :: elem
integer, intent(in), optional :: dim
integer :: size_elem_oldsoln
integer :: copydim
if (associated_elem_oldsoln(grid,elem)) then
   if (present(elem)) then
      if (present(dim)) then
         copydim = dim
         size_elem_oldsoln = size(grid%elem_oldsoln,dim=copydim)
      else
         size_elem_oldsoln = size(grid%elem_oldsoln,dim=BASIS_DIM) * &
                             size(grid%elem_oldsoln,dim=COMP_DIM) * &
                             size(grid%elem_oldsoln,dim=EIGEN_DIM)
      endif
   else
      if (present(dim)) then
         copydim = dim
         size_elem_oldsoln = size(grid%elem_oldsoln,dim=copydim)
      else
         size_elem_oldsoln = size(grid%elem_oldsoln)
      endif
   endif
else
   size_elem_oldsoln = 0
endif
end function size_elem_oldsoln

end module gridtype_mod
